#-*- coding:utf-8 -*-

#  Pybik -- A 3 dimensional magic cube game.
#  Copyright © 2009-2013  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Ported from GNUbik
# Original filename: widget-set.c
# Original copyright and license: 2003, 2004  John Darrington, GPL3+



from collections import OrderedDict

# pylint: disable=W0614,W0401
from PyQt4.QtCore import *
from PyQt4.QtCore import pyqtSignal as Signal, pyqtSlot as Slot
from PyQt4.QtGui import *

from .debug import *
# pylint: enable=W0614,W0401
from . import config
from . import dialogs
from . import plugins
from .model import models
from . import gamestate
from . import drawingarea
from .settings import settings

try:
    _, ngettext
except NameError:
    debug('gettext not found')
    _ = lambda string: string
    ngettext = lambda singular, plural, cnt: singular if cnt == 1 else plural
    
    
class AnimType (object): # pylint: disable=R0903
    NONE, FWD, BWD, KEY, NEW = list(range(5))
    
    
class ElideLabel (QLabel):
    def paintEvent(self, unused_event):
        p = QPainter(self)
        fm = QFontMetrics(self.font())
        rect = self.contentsRect()
        if fm.width(self.text()) > rect.width():
            gradient = QLinearGradient(rect.topLeft(), rect.topRight())
            start = (rect.width() - 2*rect.height()) / rect.width()
            gradient.setColorAt(start, self.palette().color(QPalette.WindowText))
            gradient.setColorAt(1.0, self.palette().color(QPalette.Window))
            pen = QPen()
            pen.setBrush(QBrush(gradient))
            p.setPen(pen)
        p.drawText(self.rect(), Qt.TextSingleLine, self.text())
        
        
class QMoveEdit (QLineEdit):
    nextword = Signal()
    prevword = Signal()
    swapnext = Signal()
    swapprev = Signal()
    
    def keyPressEvent(self, event):
        if event.matches(QKeySequence.MoveToNextWord):
            self.nextword.emit()
        elif event.matches(QKeySequence.MoveToPreviousWord):
            self.prevword.emit()
        elif event.key() == Qt.Key_Right and event.modifiers() == Qt.AltModifier:
            self.swapnext.emit()
        elif event.key() == Qt.Key_Left and event.modifiers() == Qt.AltModifier:
            self.swapprev.emit()
        else:
            QLineEdit.keyPressEvent(self, event)
            
        
class MainWindow (QMainWindow):
    
    ### init ###
    
    def __init__(self, opts):
        QMainWindow.__init__(self)
        self.ui = dialogs.UI(self, 'main.ui')
        from .ui.main import retranslate
        retranslate(self)
        
        # UI
        self.ui.splitter.setCollapsible(0, False)
        self.ui.splitter.setStretchFactor(0, 1)
        self.ui.splitter.setStretchFactor(1, 0)
        self.ui.label_debug_text.setVisible(DEBUG_DRAW or DEBUG_FPS)
        
        # set icons
        self.setWindowIcon(QIcon(config.IMAGE_FILE))
        self.ui.button_edit_exec.setIcon(QIcon.fromTheme('system-run'))
        self.ui.button_edit_clear.setIcon(QIcon.fromTheme('edit-clear'))
        self.ui.action_new.setIcon(QIcon.fromTheme('document-new'))
        self.ui.action_new_solved.setIcon(QIcon.fromTheme('document-new'))
        #self.ui.action_selectmodel.setIcon(QIcon.fromTheme('document-properties'))
        self.ui.action_quit.setIcon(QIcon.fromTheme('application-exit'))
        self.ui.action_rewind.setIcon(QIcon.fromTheme('media-seek-backward'))
        self.ui.action_previous.setIcon(QIcon.fromTheme('media-skip-backward'))
        self.ui.action_stop.setIcon(QIcon.fromTheme('media-playback-stop'))
        self.ui.action_play.setIcon(QIcon.fromTheme('media-playback-start'))
        self.ui.action_next.setIcon(QIcon.fromTheme('media-skip-forward'))
        self.ui.action_forward.setIcon(QIcon.fromTheme('media-seek-forward'))
        self.ui.action_add_mark.setIcon(QIcon.fromTheme('list-add'))
        self.ui.action_remove_mark.setIcon(QIcon.fromTheme('list-remove'))
        self.ui.action_preferences.setIcon(QIcon.fromTheme('document-properties'))
        self.ui.action_info.setIcon(QIcon.fromTheme('help-about'))
        self.ui.action_selectmodel.setShortcut(QKeySequence(settings.action.selectmodel))
        self.ui.action_initial_state.setShortcut(QKeySequence(settings.action.initial_state))
        self.ui.action_reset_rotation.setShortcut(QKeySequence(settings.action.reset_rotation))
        self.ui.action_invert_moves.setShortcut(QKeySequence(settings.action.invert_moves))
        self.ui.action_reload_scripts.setShortcut(QKeySequence(settings.action.reload_scripts))
        self.ui.action_preferences.setShortcut(QKeySequence(settings.action.preferences))
        self.ui.action_normalize_complete_rotations.setShortcut(
                                        QKeySequence(settings.action.normalize_complete_rotations))
        
        # widgets that are not created with Qt Designer
        self.playbarstate = self.create_toolbar()
        self.edit_moves = QMoveEdit()
        self.edit_moves.setObjectName("edit_moves")
        self.edit_moves.setFrame(False)
        self.ui.layout_moves.insertWidget(1, self.edit_moves)
        self.edit_moves.returnPressed.connect(self.on_edit_moves_returnPressed)
        self.edit_moves.nextword.connect(self.on_edit_moves_nextword)
        self.edit_moves.prevword.connect(self.on_edit_moves_prevword)
        self.edit_moves.swapprev.connect(self.on_edit_moves_swapprev)
        self.edit_moves.swapnext.connect(self.on_edit_moves_swapnext)
        self.cube_area = self.create_cube_area(opts)
        self.cube_area.drop_color.connect(self.on_cubearea_drop_color)
        self.cube_area.drop_file.connect(self.on_cubearea_drop_file)
        self.cube_area.debug_text.connect(self.on_cubearea_debug_text)
        self.setTabOrder(self.edit_moves, self.cube_area)
        self.setTabOrder(self.cube_area, self.ui.box_sidepane)
        self.cube_area.setFocus(Qt.OtherFocusReason)
        self.status_text = ElideLabel()
        p = self.status_text.sizePolicy()
        p.setHorizontalStretch(1)
        p.setHorizontalPolicy(QSizePolicy.Ignored)
        self.status_text.setSizePolicy(p)
        self.ui.statusbar.addWidget(self.status_text)
        # created when needed
        self.progress_dialog = None
        
        # actions that belongs to no widget
        self.action_jump_to_editbar = QAction(self)
        self.action_jump_to_editbar.setObjectName('action_jump_to_editbar')
        self.action_jump_to_editbar.setShortcut(QKeySequence(settings.action.edit_moves))
        self.action_jump_to_editbar.triggered.connect(self.on_action_jump_to_editbar_triggered)
        self.addAction(self.action_jump_to_editbar)
        self.action_edit_model = QAction(self)
        self.action_edit_model.setObjectName('action_edit_model')
        self.action_edit_model.setShortcut(QKeySequence(settings.action.edit_model))
        self.action_edit_model.triggered.connect(self.on_action_edit_model_triggered)
        self.addAction(self.action_edit_model)
        self.action_reload_shader = QAction(self)
        self.action_reload_shader.setObjectName('action_reload_shader')
        self.action_reload_shader.setShortcut(QKeySequence(settings.action.reload_shader))
        self.action_reload_shader.triggered.connect(self.on_action_reload_shader_triggered)
        self.addAction(self.action_reload_shader)
        
        # plugins
        self.treestore = QStandardItemModel()
        self.selected_tree_path = None
        self.active_script_group = 0
        self.plugin_helper = plugins.PluginHelper()
        self.add_scripts_to_sidepane(False)
        QTimer.singleShot(0, self.add_scripts_to_sidepane)
        
        self.unsolved = False
        self.animtype = AnimType.NONE
        
        # Load window state from settings
        self.ui.action_toolbar.setChecked(settings.window.toolbar)
        self.ui.action_editbar.setChecked(settings.window.editbar)
        self.ui.action_statusbar.setChecked(settings.window.statusbar)
        self.ui.toolbar.setVisible(settings.window.toolbar)
        self.ui.frame_editbar.setVisible(settings.window.editbar)
        self.ui.statusbar.setVisible(settings.window.statusbar)
        self.move_keys = self.get_move_key_dict()
        
        settings.keystore.changed.connect(self.on_settings_changed, Qt.QueuedConnection)
        settings.keystore.error.connect(self.on_settings_error)
        
        # Reload last game
        self.gamestate = gamestate.GameState()
        QTimer.singleShot(0, self.load_game)
        self.update_ui()
        
        self.resize(*settings.window.size)
        divider_position = settings.window.divider
        self.show()
        sizes = self.ui.splitter.sizes()
        self.ui.splitter.setSizes([divider_position, sum(sizes)-divider_position])
        
    def create_cube_area(self, opts):
        if DEBUG_THREAD:
            cube_area = drawingarea.CubeAreaThreaded(opts)
        else:
            cube_area = drawingarea.CubeArea(opts)
        cube_area.setObjectName('drawingarea')
        sizePolicy = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        cube_area.setSizePolicy(sizePolicy)
        self.ui.verticalLayout.addWidget(cube_area)
        cube_area.animation_ending.connect(self.on_cubearea_animation_ending)
        cube_area.request_rotation.connect(self.on_cubearea_request_rotation)
        cube_area.request_swap_blocks.connect(self.on_cubearea_request_swap_blocks)
        cube_area.request_rotate_block.connect(self.on_cubearea_request_rotate_block)
        glformat = cube_area.format()
        dialogs.PreferencesDialog.sample_buffers = max(glformat.samples(), 1)
        cube_area.start()
        return cube_area
        
    def create_toolbar(self):
        play_actions = [self.ui.action_rewind, self.ui.action_previous,
                        self.ui.action_stop, self.ui.action_play, self.ui.action_next,
                        self.ui.action_forward, self.ui.action_add_mark, self.ui.action_remove_mark]
        for action in play_actions:
            self.ui.toolbar.addAction(action)
        playbarstate = PlaybarState(play_actions)
        return playbarstate
        
    ### helper functions ###
    
    @staticmethod
    def keyval_from_name(keystr):
        key = QKeySequence(keystr)
        if key.count() == 0:
            return 0
        key = key[0]
        if 'KP' in keystr.split('+'):
            key |= int(Qt.KeypadModifier)
        return key
        
    def get_move_key_dict(self):
        return {self.keyval_from_name(key): move for (move, key) in settings.draw.accels}
        
    def add_scripts_to_sidepane(self, load=True):
        self.ui.treeview.setModel(None)
        self.treestore.clear()
        for widget in self.ui.box_sidepane.children():
            if isinstance(widget, QPushButton):
                self.ui.layout_sidepane.removeWidget(widget)
                widget.deleteLater()
        if load:
            algorithms = self.plugin_helper.load_plugins()
            if self.plugin_helper.error_messages:
                self.error_dialog('\n\n'.join(self.plugin_helper.error_messages))
        else:
            algorithms = []
        ptree = OrderedDict()
        for plugin_path, func in algorithms:
            subtree = [None, ptree]
            for name in plugin_path:
                subtree = subtree[1].setdefault(name, [None, OrderedDict()])
            if plugin_path:
                subtree[0] = func
        def fill_treeview(parent, tree):
            for name, [func, subtree] in tree.items():
                item = QStandardItem(_(name))
                item.setData(func)
                parent.appendRow(item)
                fill_treeview(item, subtree)
        for i, (name, [func, subtree]) in enumerate(ptree.items()):
            button = QPushButton(self.ui.box_sidepane)
            button.setText(_(name))
            button.setFlat(True)
            button.setFocusPolicy(Qt.TabFocus)
            self.ui.layout_sidepane.addWidget(button)
            obsc = lambda i: lambda: self.on_button_sidepane_clicked(i)
            button.clicked.connect(obsc(i))
        fill_treeview(self.treestore, ptree)
        self.set_active_script_group(self.active_script_group)
        
    def set_active_script_group(self, i):
        self.ui.layout_sidepane.insertWidget(i+1, self.ui.treeview)
        self.ui.treeview.setModel(self.treestore)
        index = self.treestore.index(i, 0)
        self.ui.treeview.setRootIndex(index)
        self.active_script_group = i
        
    def load_game(self):
        mtype, size, blocks = settings.game.type, settings.game.size, settings.game.blocks
        Model = models[mtype]
        mirror_distance = settings.draw.mirror_distance if settings.draw.mirror_faces else None
        model = Model(size, mirror_distance)
        self.gamestate.set_state(model, blocks, settings.game.moves, settings.game.position)
        self.cube_area.set_model(model, self.gamestate.current_cube_state.blocks)
        self.update_ui()
        
    def new_game(self, mtype=None, size=None, solved=False):
        if self.is_animating():
            return
        if mtype is None:
            model = self.gamestate.current_cube_state.model
            if size is None:
                size = model.sizes
            Model = model.__class__
        else:
            Model = models[mtype]
            assert size is not None
        mirror_distance = settings.draw.mirror_distance if settings.draw.mirror_faces else None
        model = Model(size, mirror_distance)
        self.gamestate.new_game(model, solved)
        self.unsolved = not solved
        self.cube_area.set_model(model, self.gamestate.current_cube_state.blocks)
        self.update_ui()
        
    def is_animating(self):
        return self.cube_area.animation_active
        
    def start_animation(self, move_data, stop_after, animtype):
        if move_data:
            blocks = self.gamestate.current_cube_state.identify_rotation_blocks(
                                move_data.axis, move_data.slice)
            self.cube_area.animate_rotation(move_data, blocks, stop_after)
            self.playbarstate.playing(self.gamestate.mark_before)
            self.animtype = animtype
            return True
        return False
        
    def update_statusbar(self):
        '''This function must be called before any action that change the move queue length
        to reflect total count of moves and after each single move'''
        
        if self.cube_area.editing_model:
            mesg = _('Press the Esc key to exit Edit Mode')
            self.unsolved = False
        else:
            current = self.gamestate.all_moves.current_place
            total = self.gamestate.all_moves.queue_length
            solved = self.gamestate.current_cube_state.is_solved()
            
            model_text = str(self.gamestate.current_cube_state.model)
            # substitution for {move_text} in statusbar text
            move_text = ngettext("{current} / {total} move",
                                 "{current} / {total} moves",
                                 total).format(current=current, total=total)
            # substitution for {solved_text} in statusbar text
            solved_text = _('solved') if solved else _('not solved')
            
            # statusbar text
            mesg = _('{model_text}, {move_text}, {solved_text}').format(
                            model_text=model_text,
                            move_text=move_text,
                            solved_text=solved_text)
            if self.unsolved and solved:
                self.unsolved = False
                QTimer.singleShot(0, self.show_solved_message)
        self.status_text.setText(mesg)
        
    def update_ui(self):
        # update toolbar
        if self.gamestate.all_moves.at_start():
            if self.gamestate.all_moves.at_end():
                self.playbarstate.empty()
            else:
                self.playbarstate.first()
        else:
            if self.gamestate.all_moves.at_end():
                self.playbarstate.last(self.gamestate.all_moves.is_mark_after(-1))
            else:
                self.playbarstate.mid(self.gamestate.all_moves.is_mark_current())
                
        # update formula
        code, pos, markup = self.gamestate.all_moves.format(
                                    self.gamestate.current_cube_state.model)
        self.set_edit_moves(code, pos, markup)
        
        self.update_statusbar()
        
    def show_solved_message(self):
        dialog = QMessageBox(self)
        dialog.setWindowTitle(_(config.APPNAME))
        mtype = _(self.gamestate.current_cube_state.model.type)
        dialog.setText(_('Congratulations, you have solved the puzzle!'))
        dialog.setIcon(QMessageBox.Information)
        dialog.setStandardButtons(QMessageBox.Close)
        self.cube_area.set_std_cursor()
        dialog.exec_()
        
    def error_dialog(self, message):
        '''Popup an error dialog box'''
        dialog = QMessageBox(self)
        dialog.setWindowTitle(_(config.APPNAME))
        dialog.setText(message)
        dialog.setIcon(QMessageBox.Warning)
        dialog.setStandardButtons(QMessageBox.Close)
        self.cube_area.set_std_cursor()
        dialog.exec_()
        
    def set_progress(self, step, message=None, value_max=None):
        if self.progress_dialog is None:
            self.progress_dialog = dialogs.ProgressDialog(self)
        canceled = self.progress_dialog.tick(step, message, value_max)
        qApp.processEvents()
        return canceled
        
    def end_progress(self):
        if self.progress_dialog is not None:
            self.progress_dialog.done()
            
    ### application handlers ###
    
    def resizeEvent(self, event):
        settings.window.size = event.size().width(), event.size().height()
        sizes = self.ui.splitter.sizes()
        if sum(sizes) > 0:  # ignore the first resize event where sizes==[0,0]
            settings.window.divider = sizes[0]
        QMainWindow.resizeEvent(self, event)
        
    def closeEvent(self, event):
        try:
            dialogs.Dialog.delete()
            self.cube_area.stop()
            model, blocks, moves, position = self.gamestate.get_state()
            try:
                mtype = models.index(model.__class__)
            except ValueError:
                print('unknown model')
            else:
                settings.keystore.changed.disconnect(self.on_settings_changed)
                settings.keystore.error.disconnect(self.on_settings_error)
                settings.game.blocks = blocks
                settings.game.moves = moves
                settings.game.position = position
                settings.game.type = mtype
                settings.game.size = model.sizes
            finally:
                settings.close()
            # Needed for PyQt4 to prevent warning:
            # QObject::startTimer: QTimer can only be used with threads started with QThread
            self.ui.treeview.setModel(None)
        finally:
            QMainWindow.closeEvent(self, event)
        
    def on_settings_changed(self, key):
        if key in ['draw.mirror_faces', 'draw.mirror_distance']:
            sizes = self.gamestate.initial_cube_state.model.sizes
            Model = self.gamestate.initial_cube_state.model.__class__
            mirror_distance = settings.draw.mirror_distance if settings.draw.mirror_faces else None
            model = Model(sizes, mirror_distance)
            self.gamestate.set_model(model)
            self.cube_area.set_model(model)
        elif key == 'draw.accels':
            self.move_keys = self.get_move_key_dict()
            
    def on_settings_error(self, message):
        self.status_text.setText(message)
        
    MODIFIER_MASK = int(Qt.ShiftModifier | Qt.ControlModifier | Qt.KeypadModifier)
    def keyPressEvent(self, event):
        if event.key() == Qt.Key_Escape:
            self.cube_area.set_editing_mode(False)
            self.update_ui()
            return
        try:
            move = self.move_keys[event.key() | int(event.modifiers()) & self.MODIFIER_MASK]
        except KeyError:
            event.ignore()
            QMainWindow.keyPressEvent(self, event)
            return
        else:
            if self.is_animating() and self.animtype != AnimType.KEY:
                return
            if not self.is_animating():
                self.gamestate.all_moves.truncate()
            self.gamestate.all_moves.parse(move, len(move), self.gamestate.current_cube_state.model)
            self.update_ui()
            if not self.is_animating():
                move_data = self.gamestate.request_next()
                self.start_animation(move_data, stop_after=False, animtype=AnimType.KEY)
            
    ### GL area handlers ###
    
    def on_cubearea_animation_ending(self, last_move):
        self.cube_area.set_transformations(self.gamestate.current_cube_state)
        
        if not last_move:
            # go again
            self.update_statusbar()
            if self.animtype == AnimType.BWD:
                move_data = self.gamestate.request_back()
            else:
                move_data = self.gamestate.request_next()
            if self.start_animation(move_data, stop_after=False, animtype=self.animtype):
                return
        self.cube_area.animation_end()
        self.update_ui()
            
    def on_cubearea_request_rotation(self, maxis, mslice, mdir):
        self.gamestate.request_rotation(maxis, mslice, mdir)
        self.update_statusbar()
        move_data = self.gamestate.request_next()
        self.start_animation(move_data, stop_after=False, animtype=AnimType.NEW)
        
    def on_cubearea_request_swap_blocks(self, blockpos, maxis, mslice, mdir):
        self.gamestate.swap_block(blockpos, maxis, mslice, mdir)
        self.cube_area.set_transformations(self.gamestate.current_cube_state)
        
    def on_cubearea_request_rotate_block(self, blockpos, rdir):
        self.gamestate.rotate_block(blockpos, rdir)
        self.cube_area.set_transformations(self.gamestate.current_cube_state)
        
    def on_cubearea_drop_color(self, blocknum, facesym, colorname):
        if blocknum < 0:
            settings.theme.bgcolor = colorname
        else:
            colornum = self.gamestate.current_cube_state.get_colornum(blocknum, facesym)
            settings.theme.face[colornum].color = colorname
            
    def on_cubearea_drop_file(self, blocknum, facesym, filename):
        #TODO: background images
        colornum = self.gamestate.current_cube_state.get_colornum(blocknum, facesym)
        settings.theme.face[colornum].image = filename
        
    def on_cubearea_debug_text(self, text):
        self.ui.label_debug_text.setText(text)
        
    ###
    
    @Slot(int, int)
    def on_splitter_splitterMoved(self, pos, index):    # pylint: disable=R0201
        if index == 1:
            settings.window.divider = pos
            
    ### sidepane handlers ###
    
    def on_button_sidepane_clicked(self, i):
        self.set_active_script_group(i)
        
    @Slot(QModelIndex)
    def on_treeview_activated(self, index):
        if self.is_animating():
            return
        func = index.data(Qt.UserRole + 1)
        if func is None:
            return
        state, moves = self.plugin_helper.call(self, func)
        if state is None:
            self.unsolved = False
        else:
            self.gamestate.initial_cube_state = state
            self.gamestate.current_cube_state = state.copy()
            self.gamestate.all_moves.reset()
            self.unsolved = True
        position = self.gamestate.all_moves.mark_and_extend(moves)
        if position >= 0:
            self.gamestate.do_fast_forward(position)
            self.cube_area.set_transformations(self.gamestate.current_cube_state)
            self.update_ui()
            move_data = self.gamestate.request_next()
            self.start_animation(move_data, stop_after=False, animtype=AnimType.NEW)
        elif state is not None:
            self.cube_area.set_transformations(self.gamestate.current_cube_state)
            self.update_ui()
        self.cube_area.update()
        
    ### action handlers ###
    
    @Slot()
    def on_action_new_activated(self):
        self.new_game(solved=False)
    @Slot()
    def on_action_new_solved_activated(self):
        self.new_game(solved=True)
    @Slot()
    def on_action_selectmodel_activated(self):
        dialog = dialogs.SelectModelDialog.run(self)
        if dialog:
            def response_ok(mtype, size, solved):
                self.new_game(mtype, size, solved)
            dialog.response_ok.connect(response_ok)
    @Slot()
    def on_action_quit_activated(self):
        self.close()
    @Slot()
    def on_action_preferences_activated(self):
        dialogs.PreferencesDialog.run(self)
    @Slot(bool)
    def on_action_toolbar_toggled(self, checked):
        self.ui.toolbar.setVisible(checked)
        settings.window.toolbar = checked
    @Slot(bool)
    def on_action_editbar_toggled(self, checked):
        self.ui.frame_editbar.setVisible(checked)
        settings.window.editbar = checked
    @Slot(bool)
    def on_action_statusbar_toggled(self, checked):
        self.ui.statusbar.setVisible(checked)
        settings.window.statusbar = checked
    @Slot()
    def on_action_reset_rotation_activated(self):
        self.cube_area.reset_rotation()
    @Slot()
    def on_action_info_activated(self):
        about = dialogs.AboutDialog(self)
        about.run()
        
    @Slot()
    def on_action_rewind_activated(self):
        if self.is_animating():
            self.cube_area.stop_requested = True
            if self.animtype == AnimType.BWD:
                self.gamestate.request_next() # undo the already applied move
            self.gamestate.do_rewind_to_mark()
            self.cube_area.animate_abort()
        else:
            needs_update = self.gamestate.do_rewind_to_mark()
            self.cube_area.set_transformations(self.gamestate.current_cube_state)
            if needs_update:
                self.cube_area.update()
                self.update_ui()
    @Slot()
    def on_action_previous_activated(self):
        if self.is_animating():
            if self.animtype == AnimType.BWD:
                # request another BWD move
                self.cube_area.stop_requested = False
            else:
                self.cube_area.stop_requested = True
                self.gamestate.request_back()
            self.cube_area.animate_abort()
            self.cube_area.stop_requested = True
        else:
            move_data = self.gamestate.request_back()
            self.start_animation(move_data, stop_after=True, animtype=AnimType.BWD)
    @Slot()
    def on_action_stop_activated(self):
        if self.is_animating():
            if self.cube_area.stop_requested:
                self.cube_area.animate_abort()
            else:
                self.cube_area.stop_requested = True
    @Slot()
    def on_action_play_activated(self):
        if not self.is_animating():
            move_data = self.gamestate.request_next()
            self.start_animation(move_data, stop_after=False, animtype=AnimType.FWD)
    @Slot()
    def on_action_next_activated(self):
        if self.is_animating():
            sr = self.cube_area.stop_requested
            if self.animtype == AnimType.BWD:
                self.cube_area.stop_requested = True
                self.gamestate.request_next()
            else:
                self.cube_area.stop_requested = False
            self.cube_area.animate_abort()
            self.cube_area.stop_requested = sr
        else:
            move_data = self.gamestate.request_next()
            self.start_animation(move_data, stop_after=True, animtype=AnimType.FWD)
    @Slot()
    def on_action_forward_activated(self):
        if self.is_animating():
            self.cube_area.stop_requested = True
            if self.animtype != AnimType.BWD:
                self.gamestate.request_back() # undo the already applied move
            self.gamestate.do_fast_forward()
            self.cube_area.animate_abort()
        else:
            self.gamestate.do_fast_forward()
            self.cube_area.set_transformations(self.gamestate.current_cube_state)
            self.cube_area.update()
            self.update_ui()
    @Slot()
    def on_action_add_mark_activated(self):
        self.gamestate.all_moves.mark_current(True)
        self.update_ui()
    @Slot()
    def on_action_remove_mark_activated(self):
        self.gamestate.all_moves.mark_current(False)
        self.update_ui()
        
    @Slot()
    def on_action_initial_state_activated(self):
        if self.is_animating():
            return
        self.gamestate.set_as_initial_state()
        self.update_ui()
    @Slot()
    def on_action_reload_scripts_activated(self):
        self.add_scripts_to_sidepane()
    @Slot()
    def on_action_invert_moves_activated(self):
        if self.is_animating():
            return
        self.gamestate.invert_moves()
        self.cube_area.set_transformations(self.gamestate.current_cube_state)
        self.cube_area.update()
        self.update_ui()
    @Slot()
    def on_action_normalize_complete_rotations_activated(self):
        if self.is_animating():
            return
        self.gamestate.normalize_complete_rotations()
        self.cube_area.set_transformations(self.gamestate.current_cube_state)
        self.cube_area.update()
        self.update_ui()
        
    ### edit formula handlers ###
    
    def on_action_jump_to_editbar_triggered(self):
        self.edit_moves.setFocus()
        
    def on_edit_moves_returnPressed(self):
        if self.is_animating():
            self.cube_area.animate_abort(update=False)
        code = self.edit_moves.text()
        pos = self.edit_moves.cursorPosition()
        self.gamestate.set_from_formula(code, pos)
        self.cube_area.set_transformations(self.gamestate.current_cube_state)
        self.cube_area.update()
        self.update_ui()
        
    def on_edit_moves_nextword(self):
        code = self.edit_moves.text()
        pos = self.edit_moves.cursorPosition()
        moves, mpos, cpos = self.gamestate.all_moves.converter.parse(code, pos,
                                        self.gamestate.current_cube_state.model)
        moves.current_place = mpos
        if cpos <= pos:
            moves.advance()
        code, pos, unused_markup = moves.format(self.gamestate.current_cube_state.model)
        self.edit_moves.setCursorPosition(pos)
        
    def on_edit_moves_prevword(self):
        code = self.edit_moves.text()
        pos = self.edit_moves.cursorPosition()
        moves, mpos, cpos = self.gamestate.all_moves.converter.parse(code, pos,
                                        self.gamestate.current_cube_state.model)
        moves.current_place = mpos
        if cpos >= pos:
            moves.retard()
        code, pos, unused_markup = moves.format(self.gamestate.current_cube_state.model)
        self.edit_moves.setCursorPosition(pos)
        
    def on_edit_moves_swapnext(self):
        code = self.edit_moves.text()
        pos = self.edit_moves.cursorPosition()
        moves, mpos, cpos = self.gamestate.all_moves.converter.parse(code, pos,
                                        self.gamestate.current_cube_state.model)
        moves.current_place = mpos
        if pos < cpos:
            moves.retard()
        moves.swapnext()
        code, pos, unused_markup = moves.format(self.gamestate.current_cube_state.model)
        self.edit_moves.setText(code)
        self.edit_moves.setCursorPosition(pos)
        
    def on_edit_moves_swapprev(self):
        code = self.edit_moves.text()
        pos = self.edit_moves.cursorPosition()
        moves, mpos, cpos = self.gamestate.all_moves.converter.parse(code, pos,
                                        self.gamestate.current_cube_state.model)
        moves.current_place = mpos
        if pos < cpos:
            moves.retard()
        moves.swapprev()
        code, pos, unused_markup = moves.format(self.gamestate.current_cube_state.model)
        self.edit_moves.setText(code)
        self.edit_moves.setCursorPosition(pos)
        
    @Slot()
    def on_button_edit_clear_clicked(self):
        self.edit_moves.setText('')
        self.on_edit_moves_returnPressed()
        
    @Slot()
    def on_button_edit_exec_clicked(self):
        self.on_edit_moves_returnPressed()
        
    def set_edit_moves(self, code, pos, unused_markup=None):
        self.edit_moves.setText(code)
        self.edit_moves.setCursorPosition(pos)
        
    def on_action_edit_model_triggered(self):
        self.cube_area.animate_abort(update=False)
        self.gamestate.do_rewind_to_start()
        if DEBUG_ROTATE:
            self.gamestate.current_cube_state.debug_blocksymbols(allsyms=True)
        self.cube_area.set_transformations(self.gamestate.current_cube_state)
        self.cube_area.set_editing_mode(True)
        self.update_ui()
        
    def on_action_reload_shader_triggered(self):
        self.cube_area.reload_shader()
        
class PlaybarState (object):
    def __init__(self, play_actions):
        self.play_button_list = play_actions
        self.empty()
        
    def set_toolbar_state(self, sflags, vflags, mark):
        if mark:
            vflags ^= 0b11
        for a in reversed(self.play_button_list):
            a.setEnabled(sflags & 1)
            a.setVisible(vflags & 1)
            sflags >>= 1
            vflags >>= 1
            
    # pylint: disable=C0321
    def empty(self):          self.set_toolbar_state(0b00000000, 0b11011101, False)
    def first(self):          self.set_toolbar_state(0b00011100, 0b11011101, False)
    def mid(self, mark):      self.set_toolbar_state(0b11011111, 0b11011110, mark)
    def last(self, mark):     self.set_toolbar_state(0b11000011, 0b11011110, mark)
    def playing(self, mark):  self.set_toolbar_state(0b11101100, 0b11101110, mark)
    # pylint: enable=C0321
    




#!/usr/bin/python3
# -*- coding: utf-8 -*-

#  Copyright © 2012-2013  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.



import os
from math import sin, cos, pi, sqrt
import re
import pickle as pickle
from collections import OrderedDict
import array

import numpy as np

from .debug import DEBUG, DEBUG_NOLABEL, DEBUG_NOBEVEL, DEBUG_INNER, DEBUG_DRAW
from .config import MODELS_DIR

N_ = lambda t: t
try:
    _
except NameError:
    _ = N_

use_modeldata = True
modeldata_cache = {}


class Face (object):
    def __init__(self, symbol, vindices=None, vertices=None):
        self.symbol = symbol
        self.vindices = vindices
        self.vertices = vertices
        self.normal = None
        self.scaled = None # Face with indices for beveled_vertices
        self.faces = None # adjacent faces
        
    def __repr__(self):
        return 'Face({}, {})'.format(self.symbol, self.vindices)
        
    def edges(self):
        vertices = self.vertices if self.vindices is None else self.vindices
        v1 = vertices[0]
        for v2 in vertices[1:]:
            yield v1, v2
            v1 = v2
        yield v1, vertices[0]
        
    def center(self, vertices=None):
        if vertices is None:
            vertices = self.vertices
        if self.vindices is not None:
            vertices = [vertices[vi] for vi in self.vindices]
        return [sum(_i)/len(_i) for _i in zip(*vertices)]
        
    def create_scaled(self, vertices, factor, new_vertices):
        bf = []
        for vi in self.vindices:
            # Scale orthogonal to the center
            bv = [(v-c)*factor+c for v, c in zip(vertices[vi], self.center(vertices))]
            bf.append(len(new_vertices))
            new_vertices.append(bv)
        self.scaled = Face(self.symbol, bf)
        
    def translated(self, vertices, offset, factor=1):
        tf = []
        if vertices is None:
            for v in self.vertices:
                tv = [xv+xo*factor for xv, xo in zip(v, offset)]
                tf.append(tv)
        else:
            for vi in self.vindices:
                tv = [xv+xo*factor for xv, xo in zip(vertices[vi], offset)]
                tf.append(tv)
        return Face(self.symbol, vertices=tf)
        
    def roll(self, n):
        if self.vindices is not None:
            vlen = len(self.vindices)
            self.vindices = [self.vindices[(i-n)%vlen] for i in range(vlen)]
        elif self.vertices is not None:
            vlen = len(self.vertices)
            self.vertices = [self.vertices[(i-n)%vlen] for i in range(vlen)]
            
    def init_adjacent_faces(self, faces):
        self.faces = []
        for edge in self.edges():
            edge = tuple(reversed(edge))
            for f in faces:
                for e in f.edges():
                    if e == edge:
                        break
                else:
                    continue
                self.faces.append(f.symbol)
                break
            else:
                assert False, 'No adjacent face for edge: ' + str((self, edge, faces))
        assert len(self.faces) == len(self.vindices)
        
    def flip(self):
        self.vertices.reverse()
        if self.normal is not None:
            for i in range(len(self.normal)):
                self.normal[i] = -self.normal[i]
        
        
class Block (object):
    def __init__(self, model, index, data=None):
        if data is not None:
            self.__dict__.update(data)
            return
        self.index = index
        self.indices = (index % model.sizes[0],
                        index // model.sizes[0] % model.sizes[1],
                        index // model.sizes[0] // model.sizes[1])
        self.coords = [float(2*idx - model.sizes[i] + 1) for i, idx in enumerate(self.indices)]
        self.coords.append(1.)
        self.symbol_to_move = {}
        self.visible_faces = []
        self.visible_glfaces = []
        for symbol in model.faces:
            mdir = symbol not in model.symbols
            maxis = model.symbols.index(model.invers_symbols[symbol] if mdir else symbol)
            mslice = self.axis_to_slice(maxis)
            self.symbol_to_move[symbol] = (maxis, mslice, mdir)
            if mslice == (model.sizes[maxis]-1 if mdir else 0):
                self.visible_faces.append(symbol)
        self.inplace_rotations = []
        
    def init_inplace_rotations(self, model):
        for (blocknum, rotsym), blocknum2 in model.rotated_position.items():
            if blocknum == blocknum2 == self.index and rotsym:
                self.inplace_rotations.append(rotsym)
        self.inplace_rotations = sorted(self.inplace_rotations, key=len)[:2]
        
    def __repr__(self):
        return 'Block({}, {}, {}, {})'.format(self.index, self.indices, self.coords, self.visible_faces)
        
    def axis_to_slice(self, maxis):
        return self.indices[maxis]
        
        
class Geometry (object):
    def __init__(self, vertices=None, faces=None, bevel_factor=None):
        self.vertices = vertices
        self.faces = faces
        self.beveled_vertices = [] # list of vertex positions
        
        for f in self.faces.values():
            f.normal = f.center(self.vertices)
            f.create_scaled(self.vertices, bevel_factor, self.beveled_vertices)
            f.init_adjacent_faces(list(self.faces.values()))
            
    def create_permutation(self, matrix):
        m = np.matrix(matrix)
        indices = {}
        for i, v in enumerate(self.vertices):
            rv = (m * np.matrix(v+[0]).T).T.tolist()[0][:-1]
            ri = self.vertices.index(rv)
            indices[i] = ri
        symbols = {}
        for sym, f in self.faces.items():
            rvindices = [indices[i] for i in f.vindices]
            for rsym, rf in self.faces.items():
                if set(rvindices) == set(rf.vindices):
                    symbols[rsym.lower()] = sym.lower()
        return symbols
        
        
class EmptyModel (object):
    type = 'Empty'
    mformat = 'Empty'
    symbols = ''
    symbolsI = ''
    faces = ''
    
    sizes = [0, 0, 0]
    size = None
    bounding_sphere_radius = 1.
    blocks = []
    
    def __init__(self, *unused_args):
        pass
        
    def gl_vertex_data(self, unused_selection_mode, unused_design):
        return b'', b'', b'', b'', [], 0, [], 0, 0
        
        
class BrickModel (object):
    type = N_('Brick')
    mformat = _('{0}×{1}×{2}-Brick')
    symmetry = 180., 180., 180.
    
    axes = [(-1,0,0), (0,-1,0), (0,0,-1)]   # pylint: disable=C0324
    symbols = 'LDB'
    symbolsI = 'RUF'
    faces = 'UDLRFB'
    epsilon = 0.00001
    
    #### geometry of the cube ####
    #  vertex      standard
    #  indices    orientation    face symbols
    # 2------3                     +---+
    # |\     |\      y             | U |
    # | 6------7     |         +---+---+---+---+
    # | |    | |     o--x      | L | F | R | B |
    # 0-|----1 |      \        +---+---+---+---+
    #  \|     \|       z           | D |
    #   4------5                   +---+
    
    geom = Geometry(
                # vertex-positions, used for unbeveled faces and picking
                vertices=[[-1,-1,-1], [1,-1,-1], [-1, 1,-1], [1, 1,-1],
                          [-1,-1, 1], [1,-1, 1], [-1, 1, 1], [1, 1, 1]],
                # vertex-indices for unbeveled faces, used for picking
                faces=OrderedDict((
                        ('U', Face('U', [2,6,7,3])), ('D', Face('D', [4,0,1,5])),
                        ('L', Face('L', [2,0,4,6])), ('R', Face('R', [7,5,1,3])),
                        ('F', Face('F', [6,4,5,7])), ('B', Face('B', [3,1,0,2])))),
                bevel_factor=0.9
            )
    texpos_tiled = [(0, 0),  (0, 1),  (1, 1),  (1, 0)]
    face_axes = OrderedDict((
                        ('U', (0, 2)), ('D', (0, 2)),
                        ('L', (2, 1)), ('R', (2, 1)),
                        ('F', (0, 1)), ('B', (0, 1))))
    
    def __init__(self, size, mirror_distance):
        self.sizes = self.norm_sizes(size) # tuple of lenght 3
        assert len(self.sizes) == 3
        for _size in self.sizes:
            assert 1 <= _size <= 10
        self.size = size # derived classes can assign a different value
        self.mirror_distance = mirror_distance and mirror_distance * max(self.sizes)
        
        #TODO: The bounding_sphere_radius is optimised for the far clipping plane,
        #      for the near clipping plane the radius without the mirror_distance
        #      would be sufficient.
        s = max(self.sizes)
        sm = s + (self.mirror_distance or 0)
        self.bounding_sphere_radius = sqrt(2*s*s + sm*sm)
        
        self.invers_symbols = {}
        for sym, isym in zip(self.symbols, self.symbolsI):
            self.invers_symbols[sym] = isym
            self.invers_symbols[isym] = sym
        self.axesI = [tuple(-x for x in a) for a in self.axes]
        
        self.normal_rotation_symbols = {}
        self.rotation_matrices = {}
        self._init_rotations()
        self.face_permutations = self._create_permutations()
        
        if use_modeldata:
            global modeldata_cache
            try:
                modeldata = modeldata_cache[(self.type, self.sizes)]
            except KeyError:
                datafilename = self.get_datafilename(self.sizes)
                datafilename = os.path.join(MODELS_DIR, datafilename)
                with open(datafilename, 'rb') as datafile:
                    modeldata = pickle.load(datafile)
                modeldata_cache = modeldata
                modeldata = modeldata_cache[(self.type, self.sizes)]
            self.blocks = [Block(None, None, data=data) for data in modeldata['blocks']]
            self.rotated_position = modeldata['rotated_position']
            self.gl_never_visible_face = self._gl_never_visible_face_cached
        else:
            nblocks = self.sizes[0] * self.sizes[1] * self.sizes[2]
            self.blocks = [Block(self, i) for i in range(nblocks)]
            self.rotated_position = self._create_rotated_position()
            for block in self.blocks:
                block.init_inplace_rotations(self)
            self.gl_never_visible_face = self._gl_never_visible_face_create_cache
        
        self.pick_vector = {}
        for maxis, axis in enumerate(self.axes):
            for symbol, face in self.geom.faces.items():
                self.pick_vector[maxis, False, symbol] = np.cross(axis, face.normal).tolist()
                self.pick_vector[maxis, True, symbol] = np.cross(face.normal, axis).tolist()
                
        self.pick_data = [()]  # list of (maxis, mslice, mdir, face, center, block, symbol)
        
    @classmethod
    def norm_sizes(cls, sizes):
        return sizes
        
    @classmethod
    def displaystring(cls, size):
        return cls.mformat.format(*size)
        
    def __str__(self):
        return self.displaystring(self.sizes)
        
    @classmethod
    def _create_rotation(cls, axis, angle):
        angle = angle / 180. * pi
        sa = sin(angle)
        ca = cos(angle)
        e_ca = 1 - ca
        n1 = axis[0]
        n2 = axis[1]
        n3 = axis[2]
        m = np.matrix([
            [n1*n1*e_ca + ca,    n1*n2*e_ca - n3*sa, n1*n3*e_ca + n2*sa, 0.],
            [n2*n1*e_ca + n3*sa, n2*n2*e_ca + ca,    n2*n3*e_ca - n1*sa, 0.],
            [n3*n1*e_ca - n2*sa, n3*n2*e_ca + n1*sa, n3*n3*e_ca + ca,    0.],
            [0.,    0.,     0.,     1.],
        ])
        #XXX: try to keep the matrix clean
        mx, my = m.shape
        for y in range(my):
            for x in range(mx):
                if abs(m.A[y][x]) < cls.epsilon:
                    m.A[y][x] = 0.
        return m
        
    @classmethod
    def _matrix_equal(cls, m1, m2):
        assert m1.shape == m2.shape, (m1, m2)
        mx, my = m1.shape
        for y in range(my):
            for x in range(mx):
                if abs(m1.A[y][x] - m2.A[y][x]) > cls.epsilon:
                    return False
        return True
        
    def _init_rotations(self):
        prim = [(sym, self._create_rotation(axis, angle))
                    for axis, sym, angle in zip(self.axes + self.axesI,
                                         self.symbols + self.symbolsI,
                                         self.symmetry + self.symmetry)]
        self.normal_rotation_symbols[''] = ''
        transform = [('', np.matrix(np.identity(4)))]
        for sp, p in prim:
            transform.append((sp, p))
        for sm, m in transform:
            for sp, p in prim:
                n = m * p
                sn = sm + sp
                for st, t in transform:
                    if self._matrix_equal(t, n):
                        self.normal_rotation_symbols[sn] = st
                        break
                else:
                    self.normal_rotation_symbols[sn] = sn
                    transform.append((sn, n))
        for sm, m in transform:
            self.rotation_matrices[sm] = m.A.tolist()
        
    def _create_permutations(self):
        face_permutations = {}
        for msym, matrix in self.rotation_matrices.items():
            face_permutations[msym] = self.geom.create_permutation(matrix)
        return face_permutations
        
    def _create_rotated_position(self):
        rotated_position = {}
        for b, block in enumerate(self.blocks):
            for sym, rotation in self.rotation_matrices.items():
                coords = (np.matrix([block.coords]) * rotation).A.tolist()[0]
                for p, pos in enumerate(self.blocks):
                    if pos.coords == coords:
                        rotated_position[b, sym] = p
                        break
                else:
                    assert False, 'not a permutation'
        return rotated_position
        
    @classmethod
    def get_datafilename(cls, sizes):
        x, y, unused_z = sizes
        if x <= 2:
            return 'mdata_01-02'
        else:
            return 'mdata_{:02}_{}'.format(x, 0 if x<=5 else y%2 if x<=8 else y%3)
        
    def get_savedata(self):
        self.calc_data()
        blocks = [block.__dict__ for block in self.blocks]
        return {'blocks': blocks, 'rotated_position': self.rotated_position}
            
    def norm_symbol(self, sym):
        try:
            return self.normal_rotation_symbols[sym]
        except KeyError:
            new_sym = ''
            for c in sym:
                try:
                    new_sym = self.normal_rotation_symbols[new_sym+c]
                except KeyError:
                    raise ValueError('invalid symbol:', sym)
            return new_sym
            
    def block_indices_to_index(self, indices):
        indices = tuple(indices)
        for b, block in enumerate(self.blocks):
            if block.indices == indices:
                return b
        raise ValueError('Invalid block indices:', indices)
        
    def rotation_symbolic_to_matrix(self, block, sym):
        m = self.rotation_matrices[sym][:]
        m[3] = self.blocks[block].coords
        return m
        
    def block_symbolic_to_block_index(self, symblock):
        indices = [1] * len(self.axes)
        for match in re.finditer(r'(.)(\d*)', symblock):
            blockface, blockslice = match.groups()
            blockface = blockface.upper()
            blockslicenum = int(blockslice)-1 if blockslice else 0
            if blockface in self.symbolsI:
                axis = self.symbolsI.index(blockface)
                blockslicenum = self.sizes[axis]-1 - blockslicenum
                blockface = self.invers_symbols[blockface]
            else:
                axis = self.symbols.index(blockface)
            indices[axis] = blockslicenum
        return self.block_indices_to_index(indices)
        
    def block_index_to_block_symbolic(self, blockpos, rotsym):
        def axisidx_to_sym(axis, idx):
            if idx <= self.sizes[axis] // 2:
                sym = self.symbols[axis]
            else:
                idx = self.sizes[axis]-1 - idx
                sym = self.symbolsI[axis]
            sym = sym.lower()
            if idx == 0:
                # skip idx for corners
                return sym, self.face_symbolic_to_face_color(sym, rotsym)
            elif idx == 1 and self.sizes[axis] == 3:
                # for size == 3 there is only one edge
                return '', ''
            else:
                # symbol with index to distinguish edge, but no color because the face is not visible
                return sym + str(idx+1), '?'
        x, y, z = self.blocks[blockpos].indices
        symx, colorsymx = axisidx_to_sym(0, x)
        symy, colorsymy = axisidx_to_sym(1, y)
        symz, colorsymz = axisidx_to_sym(2, z)
        return symx + symy + symz, colorsymx + colorsymy + colorsymz
        
    def face_symbolic_to_face_color(self, face, rot):
        for k, v in self.face_permutations[rot].items():
            if v == face:
                return k
        else:
            assert False, (face, rot)
        
    def rotate_symbolic(self, axis, rdir, block, sym):
        rsym = (self.symbols if not rdir else self.symbolsI)[axis]
        block = self.rotated_position[block, rsym]
        sym = self.norm_symbol(sym + rsym)
        return block, sym
        
    def rotate_move(self, complete_move, move):
        caxis, unused_cslice, cdir = complete_move
        maxis, mslice, mdir = move
        caxissym = (self.symbols if cdir else self.symbolsI)[caxis]
        maxissym = (self.symbols if not mdir else self.symbolsI)[maxis]
        raxissym = self.face_permutations[caxissym][maxissym.lower()].upper()
        rdir = raxissym not in self.symbols
        raxis = self.symbols.index(self.invers_symbols[raxissym] if rdir else raxissym)
        if mdir != rdir:
            mslice = self.sizes[raxis] - 1 - mslice
        return raxis, mslice, rdir
        
    @staticmethod
    def get_selected_move(block, face, edgeno):
        '''block: a block in the rotation slice
           face -> edgeno is the direction of slice rotation
        '''
        edgeno = (edgeno - 1) % 4
        rotation_symbol = face.faces[edgeno]
        return block.symbol_to_move[rotation_symbol]
        
    def compare_move_to_pick(self, maxis, unused_mslice, mdir, face, faceedge):
        axis = self.axes[maxis]
        if mdir:
            vpick = np.cross(axis, face.normal).tolist()
        else:
            vpick = np.cross(face.normal, axis).tolist()
        #TODO: rather test whether face.center+pick_vector intersects with the edge
        return vpick == faceedge.normal
        
    @staticmethod
    def get_selected_move_center(block, face):
        maxis, mslice, mdir = block.symbol_to_move[face.symbol]
        return maxis, mslice, not mdir
        
    def _gl_never_visible_face_cached(self, block, face):   # pylint: disable=R0201
        return face not in block.visible_glfaces
        
    def _gl_never_visible_face_create_cache(self, block, face):
        if not self._gl_never_visible_face_calculated(block, face):
            block.visible_glfaces.append(face)
        # create the cache and discard the real gl face
        return True
        
    def _gl_never_visible_face_calculated(self, block, face):
        if max(self.sizes) == 2 and self.sizes.count(2) >= 2:
            #FIXME: the algorithm to detect invisible faces is wrong,
            # deactivate the test for the cubes that are most affected.
            return False
        vertices = self.geom.beveled_vertices
        c = block.coords
        sqr0 = self.sizes[0] * self.sizes[0]
        sqr1 = self.sizes[1] * self.sizes[1]
        sqr2 = self.sizes[2] * self.sizes[2]
        for vi in face:
            v = [vk+ck for vk, ck in zip(vertices[vi], c)]
            if (v[0]*v[0] + v[1]*v[1] > min(sqr0, sqr1) or
                v[0]*v[0] + v[2]*v[2] > min(sqr0, sqr2) or
                v[1]*v[1] + v[2]*v[2] > min(sqr1, sqr2)):
                return False
        return True
        
    def texpos_tiled_to_mosaic(self, block, face, texpos_tiled):
        axisx, axisy = self.face_axes[face.symbol]
        sizex, sizey = self.sizes[axisx], self.sizes[axisy]
        tptx, tpty = texpos_tiled
        
        vertices = [self.geom.vertices[vi] for vi in face.vindices]
        subx = [(c1-c0)/2 for c0, c1 in zip(vertices[0], vertices[-1])]
        suby = [(c1-c0)/2 for c0, c1 in zip(vertices[0], vertices[1])]
        coords = block.coords[:3]
        tpmx = sum(tc*bc for tc, bc in zip(subx, coords))
        tpmy = sum(tc*bc for tc, bc in zip(suby, coords))
        
        texx = ((sizex-1 + tpmx) / 2. + tptx) / sizex
        texy = ((sizey-1 + tpmy) / 2. + tpty) / sizey
        return texx, texy
        
    def gl_label_quads(self, block, visible):
        if DEBUG:
            if DEBUG_INNER:
                for i in range(3):
                    if block.indices[i] in [0, self.sizes[i]-1]:
                        return
            if DEBUG_NOLABEL and visible:
                return
            if DEBUG_NOBEVEL and not visible:
                return
        for faceno, symbol in enumerate(self.faces):
            face = self.geom.faces[symbol]
            if (symbol in block.visible_faces) != visible:
                continue
            f = face.scaled
            if self.gl_never_visible_face(block, f.vindices):
                continue
            for i in [0, 1, 2, 0, 2, 3]:
                vi = f.vindices[i]
                v = self.geom.beveled_vertices[vi]
                texpos_tiled = self.texpos_tiled[i]
                texpos_mosaic = self.texpos_tiled_to_mosaic(block, face, texpos_tiled)
                yield (v, face.normal, faceno, texpos_tiled, texpos_mosaic)
            if self.mirror_distance is not None and visible:
                f = f.translated(self.geom.beveled_vertices, face.normal, self.mirror_distance)
                f.normal = face.normal[:]
                f.flip()
                for i in [0, 1, 2, 0, 2, 3]:
                    v = f.vertices[i]
                    tptx, tpty = self.texpos_tiled[i]
                    texpos_tiled = 1. - tptx, tpty
                    texpos_mosaic = self.texpos_tiled_to_mosaic(block, face, texpos_tiled)
                    yield (v, f.normal, faceno, texpos_tiled, texpos_mosaic)
                    
    def gl_beveled_quads(self, block):
        '''For every edge create a face'''
        if DEBUG:
            if DEBUG_INNER:
                for i in range(3):
                    if block.indices[i] in [0, self.sizes[i]-1]:
                        return
            if DEBUG_NOBEVEL:
                return
        for symbol, f in self.geom.faces.items():
            for symbol2, (vi, vi2) in zip(f.faces, f.edges()):
                if symbol >= symbol2: # find the corners only once, no matter in which order
                    continue
                f2 = self.geom.faces[symbol2]
                # f and f2 have a common edge (vi,vi2)
                # we now need the vertices and normals of the adjacent labels.
                # remember, the face must be counterclockwise!
                bvi1 = f.scaled.vindices[f.vindices.index(vi2)]
                bvi2 = f.scaled.vindices[f.vindices.index(vi)]
                bvi3 = f2.scaled.vindices[f2.vindices.index(vi)]
                bvi4 = f2.scaled.vindices[f2.vindices.index(vi2)]
                if self.gl_never_visible_face(block, [bvi1, bvi2, bvi3, bvi4]):
                    continue
                yield self.geom.beveled_vertices[bvi1], f.normal
                yield self.geom.beveled_vertices[bvi2], f.normal
                yield self.geom.beveled_vertices[bvi3], f2.normal
                yield self.geom.beveled_vertices[bvi1], f.normal
                yield self.geom.beveled_vertices[bvi3], f2.normal
                yield self.geom.beveled_vertices[bvi4], f2.normal
                
    def gl_beveled_triangles(self, block):
        '''For every corner create a face'''
        if DEBUG:
            if DEBUG_INNER:
                for i in range(3):
                    if block.indices[i] in [0, self.sizes[i]-1]:
                        return
            if DEBUG_NOBEVEL:
                return
        for vi in range(len(self.geom.vertices)):
            bf = [] # one beveled face for each vertex
            bn = []
            for face_first in self.geom.faces.values():
                if vi in face_first.vindices:
                    # we now have one adjacent face, this code should be reached if the model is valid
                    break
            else:
                face_first = None
                assert False
            face = face_first
            while True:
                ivi = face.vindices.index(vi)
                bvi = face.scaled.vindices[ivi]
                bf.insert(0, bvi) # we need counterclockwise order
                bn.insert(0, face.normal)
                symbol = face.faces[ivi]
                face = self.geom.faces[symbol]
                ## we now have the clockwise next face
                if face_first.symbol == face.symbol:
                    break
            if self.gl_never_visible_face(block, bf):
                continue
            for vi, n in zip(bf, bn):
                yield self.geom.beveled_vertices[vi], n
            
    def gl_pick_triangles(self, selection_mode):
        def edge_center(v1, v2):
            return [(_v1 + _v2) / 2 for _v1, _v2 in zip(v1, v2)]
        for block in self.blocks:
            cnt_faces = 0
            for i in range(3):
                if block.indices[i] in [0, self.sizes[i]-1]:
                    cnt_faces += 1
            for face, symbol in enumerate(self.faces):
                f = self.geom.faces[symbol]
                assert f.symbol == symbol
                if symbol not in block.visible_faces:
                    continue
                cnt_n = 0
                for _sym in f.faces:
                    if _sym in block.visible_faces:
                        cnt_n += 1
                ft = f.translated(self.geom.vertices, block.coords)
                vc = ft.center()
                if self.mirror_distance is None:
                    ftt = ft
                else:
                    ftt = ft.translated(None, f.normal, self.mirror_distance)
                vct = ftt.center()
                
                if selection_mode == 1 and cnt_n == 0:
                    maxis, mslice, mdir = self.get_selected_move_center(block, f)
                    color = len(self.pick_data)
                    self.pick_data.append((maxis, mslice, mdir, face, True, block, symbol, None, None))
                    for i in (0,1,2, 2,3,0):    # pylint: disable=C0324
                        yield color, ft.vertices[i]
                    if self.mirror_distance is not None:
                        color = len(self.pick_data)
                        self.pick_data.append((maxis, mslice, not mdir, face, True, block, symbol, None, None))
                        for i in (0,2,1, 2,0,3):    # pylint: disable=C0324
                            yield color, ftt.vertices[i]
                elif selection_mode == 1 and cnt_n == 1 and cnt_faces == 2:
                    for edgeno, symboledge in enumerate(f.faces): # find the other face on the block
                        if symboledge in block.visible_faces:
                            break
                    else:
                        edgeno = symboledge = None
                        assert False
                    maxis, mslice, mdir = self.get_selected_move(block, f, edgeno)
                    if DEBUG:
                        assert self.compare_move_to_pick(maxis, mslice, mdir, f, self.geom.faces[symboledge])
                    ec = edge_center(*list(ft.edges())[edgeno])
                    color = len(self.pick_data)
                    self.pick_data.append((maxis, mslice, mdir, face, False, block, symbol, vc, ec))
                    for i in (0,1,2, 2,3,0):    # pylint: disable=C0324
                        yield color, ft.vertices[i]
                    if self.mirror_distance is not None:
                        for i in (0,2,1, 2,0,3):    # pylint: disable=C0324
                            yield color, ftt.vertices[i]
                elif selection_mode == 1 and cnt_n == 2 and cnt_faces == 3:
                    # find the two other faces on the block
                    for i1, symboledge1 in enumerate(f.faces): # find the other face on the block
                        i2 = (i1 + 1) % 4
                        symboledge2 = f.faces[i2]
                        visible_faces = block.visible_faces
                        if symboledge1 in visible_faces and symboledge2 in visible_faces:
                            break
                    else:
                        i1 = symboledge1 = None
                        assert False
                    symboledges = (i1, symboledge1, i1-1), (i2, symboledge2, i2)
                    for edgeno, symboledge, offset in symboledges:
                        maxis, mslice, mdir = self.get_selected_move(block, f, edgeno)
                        if DEBUG:
                            assert self.compare_move_to_pick(maxis, mslice, mdir, f, self.geom.faces[symboledge])
                        ec = edge_center(*list(ft.edges())[edgeno])
                        color = len(self.pick_data)
                        self.pick_data.append((maxis, mslice, mdir, face, False, block, symbol, vc, ec))
                        for i in (0, 1, 2):
                            yield color, ft.vertices[(i+offset)%4]
                        if self.mirror_distance is not None:
                            for i in (0, 2, 1):
                                yield color, ftt.vertices[(i+offset)%4]
                else:
                    for edgeno, (edge, edget) in enumerate(zip(ft.edges(), ftt.edges())):
                        maxis, mslice, mdir = self.get_selected_move(block, f, edgeno)
                        ec = edge_center(*edge)
                        color = len(self.pick_data)
                        self.pick_data.append((maxis, mslice, mdir, face, False, block, symbol, vc, ec))
                        # pylint: disable=C0321
                        yield color, vc; yield color, edge[0]; yield color, edge[1]
                        if self.mirror_distance is not None:
                            yield color, vct; yield color, edget[1]; yield color, edget[0]
                    
    def gl_vertex_data(self, selection_mode, design):
        vertices = []
        normals = []
        colors = []
        texpos = []
        # list of [count vertices, face index]
        labelinfos = [None for unused_b in self.blocks]
        cnts_block = [0 for unused_b in self.blocks]
        
        # label vertices
        for iblock, block in enumerate(self.blocks):
            cnt = 0
            facesinfo = []
            last_faceno = -1
            for v, n, f, tpt, tpm in self.gl_label_quads(block, True):
                vertices += v
                normals += n
                colors += design.label_colors[f]
                texpos += tpm if design.imagemodes[f] else tpt
                if last_faceno != f:
                    cnt = 1
                    facesinfo.append((cnt, f))
                    last_faceno = f
                else:
                    cnt += 1
                    facesinfo[-1] = (cnt, f)
            labelinfos[iblock] = facesinfo
            
        # bevel, hidden labels
        idx_block = len(vertices) // 3
        for iblock, block in enumerate(self.blocks):
            cnt = 0
            for v, n in self.gl_beveled_quads(block):
                vertices += v
                normals += n
                colors += design.bevel_color
                cnt += 1
            for v, n, f, tpt, tpm in self.gl_label_quads(block, False):
                vertices += v
                normals += n
                colors += design.bevel_color
                cnt += 1
            for v, n in self.gl_beveled_triangles(block):
                vertices += v
                normals += n
                colors += design.bevel_color
                cnt += 1
            cnts_block[iblock] = cnt
        
        pickvertices, pickcolors = self.gl_pick_data(selection_mode)
        idx_pick = len(vertices) // 3
        cnt_pick = len(pickvertices) // 3
        
        vertices += pickvertices
        colors += pickcolors
        
        idx_debug = len(vertices) // 3
        if DEBUG_DRAW:
            x = float(-self.sizes[0])
            y = float(-self.sizes[1])
            z = float(-self.sizes[2])
            # axes
            vertices += [x-1, y-1, z,  -x, y-1, z]
            colors += [255, 0, 0,  255, 0, 0]
            vertices += [x-1, y-1, z,  x-1, -y, z]
            colors += [0, 255, 0,  0, 255, 0]
            vertices += [x-1, y-1, z,  x-1, y-1, -z]
            colors += [0, 0, 255,  0, 0, 255]
            # selection (modelview)
            vertices += [0, 0, 0,  1, 1, 1]
            colors += [255, 255, 255,  255, 255, 255]
            # selection (viewport)
            vertices += [0, 0, 0,  1, 1, 0]
            colors += [255, 255, 0,  255, 0, 0]
            
        if DEBUG:
            assert sum([cnt for _bi in labelinfos for cnt, unused_faceno in _bi]) == idx_block
        
        return (    array.array('f', vertices).tobytes(),
                    array.array('f', normals).tobytes(),
                    array.array('B', colors).tobytes(),
                    array.array('f', texpos).tobytes(),
                    labelinfos,
                    idx_block, cnts_block,
                    idx_pick, cnt_pick,
                    idx_debug,
               )
                
    def calc_data(self):
        '''Function to verify that Block.visible_glfaces lists are initialized.'''
        for iblock, block in enumerate(self.blocks):
            for v, n, f, tpt, tpm in self.gl_label_quads(block, True):
                pass
            for v, n in self.gl_beveled_quads(block):
                pass
            for v, n, f, tpt, tpm in self.gl_label_quads(block, False):
                pass
            for v, n in self.gl_beveled_triangles(block):
                pass
                
    def gl_pick_data(self, selection_mode):
        # Pick TRIANGLES vertices
        vertices = []
        colors = []
        for col, v in self.gl_pick_triangles(selection_mode):
            vertices += v
            color = [(col>>4) & 0xf0, (col) & 0xf0, (col<<4) & 0xf0]
            colors += color
        assert len(vertices) == len(colors)
        return vertices, colors
        
        
class TowerModel (BrickModel):
    type = N_('Tower')
    mformat = _('{0}×{1}-Tower')
    symmetry = 180., 90., 180.
    
    def __init__(self, size, mirror_distance):
        BrickModel.__init__(self, size, mirror_distance)
        self.size = size[:2]
        
    @classmethod
    def norm_sizes(cls, sizes):
        x, y, unused_z = sizes
        return x, y, x
        
        
class CubeModel (BrickModel):
    type = N_('Cube')
    mformat = _('{0}×{0}×{0}-Cube')
    symmetry = 90., 90., 90.
    
    def __init__(self, size, mirror_distance):
        BrickModel.__init__(self, size, mirror_distance)
        self.size = size[0],
        
    @classmethod
    def norm_sizes(cls, sizes):
        x, unused_y, unused_z = sizes
        return x, x, x
        
        
empty_model = EmptyModel()
models = [CubeModel, TowerModel, BrickModel]

def from_string(modelstr):
    if modelstr == '*':
        return '*', '*', (), None
    re_model = r'''^(\w+)
                    (?: \s+(\w+)
                        (?: \s*(\W)
                            \s*(\w+)
                            (?: \s*\3
                                \s*(\w+)
                        )?)?
                        (?:\s+with
                            \s+(.+)
                    )?)?\s*$'''
    match = re.match(re_model, modelstr, re.X)
    if match is None:
        raise ValueError('Invalid model: ' + modelstr)
    mtype, width, height, depth, exp = match.group(1, 2, 4, 5, 6)
    for Model in models:
        if mtype == Model.type:
            break
    else:
        Model = None
        raise ValueError('Unknown model type %r' % mtype)
    def convert_if_int(value):
        if value is None:
            return value
        try:
            return int(value)
        except ValueError:
            return value
    sizes = tuple(convert_if_int(s) for s in (width, height, depth))
    return modelstr, Model, sizes, exp
    
    



#-*- coding:utf-8 -*-

#  Pybik -- A 3 dimensional magic cube game.
#  Copyright © 2009, 2011-2013  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Ported from GNUbik
# Original filename: menus.c
# Original copyright and license: 2003, 2004  John Darrington,  GPL3+

# pylint: disable=C0321


import os
import re

# pylint: disable=W0614,W0401
from PyQt4.QtCore import *
from PyQt4.QtCore import pyqtSignal as Signal, pyqtSlot as Slot
from PyQt4.QtGui import *

from .debug import *
# pylint: enable=W0614,W0401
from . import config
from .settings import settings
from .textures import textures
from .model import models

try:
    _
except NameError:
    _ = lambda t: t
    

class Dialog (QDialog):
    _instance = None
    _classes = []
    
    def __init__(self, parent):
        QDialog.__init__(self, parent)
        self.accepted.connect(self.on_response)
        self._ignore_changed = False
        self.bound = {}
        settings.keystore.changed.connect(self.on_settings_changed, Qt.QueuedConnection)
        
    @classmethod
    def run(cls, parent_window=None):
        if cls._instance is None:
            cls._instance = cls(parent_window)
            ret = cls._instance
            cls._classes.append(cls)
        else:
            ret = None
            cls._instance.move(cls._instance.pos())
        cls._instance.show()
        return ret
        
    @classmethod
    def delete(cls):
        while cls._classes:
            cls_ = cls._classes.pop(0)
            # pylint: disable=W0212
            cls_._instance.close()
            cls_._instance.deleteLater()
            # pylint: enable=W0212
            cls_._instance = None
            
    def bind(self, key, widget, getter, setter, signal):
        if isinstance(getter, str): getter = getattr(widget, getter)
        if isinstance(setter, str): setter = getattr(widget, setter)
        if isinstance(signal, str): signal = getattr(widget, signal)
        setter(getattr(settings, key))
        if signal is not None:
            signal.connect(lambda: self.on_widget_changed(key))
        self.bound[key] = (getter, setter, signal)
        
    @staticmethod
    def bind_reset(settings_key, button):
        button.clicked.connect(lambda: delattr(settings, settings_key))
        
    def on_response(self):
        pass
        
    def on_widget_changed(self, key):
        if self._ignore_changed:
            return
        try:
            getter = self.bound[key][0]
        except KeyError:
            return
        self._ignore_changed = True
        setattr(settings, key, getter())
        self._ignore_changed = False
        
    def on_settings_changed(self, key):
        if self._ignore_changed:
            return
        try:
            setter = self.bound[key][1]
        except KeyError:
            return
        self._ignore_changed = True
        setter(getattr(settings, key))
        self._ignore_changed = False
        
        
class ColorIconEngine (QIconEngineV2):
    def __init__(self):
        QIconEngineV2.__init__(self)
        self.color = 'black'
        
    def paint(self, painter, rect, unused_mode, unused_state):
        painter.fillRect(rect, QColor(self.color))
        
        
class ColorButton (QPushButton):
    color_changed = Signal(str)
    
    def __init__(self, replace_button):
        self._icon = ColorIconEngine()
        parent = replace_button.parentWidget()
        QPushButton.__init__(self, QIcon(self._icon), '', parent)
        height = self.iconSize().height()
        self.setIconSize(QSize(height * self.width() // self.height(), height))
        self.setSizePolicy(replace_button.sizePolicy())
        self.setObjectName(replace_button.objectName())
        
        layout = replace_button.parentWidget().layout()
        index = layout.indexOf(replace_button)
        position = layout.getItemPosition(index)
        layout.removeWidget(replace_button)
        layout.addWidget(self, *position)
        replace_button.deleteLater()
        self.clicked.connect(self.on_clicked)
        
    def set_color(self, color):
        self._icon.color = color
        self.update()
        
    def get_color(self):
        return self._icon.color
        
    def on_clicked(self):
        dialog = QColorDialog(self)
        dialog.setCurrentColor(QColor(self.get_color()))
        if dialog.exec_() == QDialog.Accepted:
            color = dialog.currentColor().name()
            self.set_color(color)
            self.color_changed.emit(color)
            
        
class ShortcutEditor (QLabel):
    key_pressed = Signal()
    
    def __init__(self, parent):
        QLabel.__init__(self, _('Press a key …'), parent)
        self.setFocusPolicy(Qt.StrongFocus)
        self.setAutoFillBackground(True)
        self.key = None
        
    SAFE_MODIFIER_MASK = Qt.ShiftModifier | Qt.ControlModifier
    IGNORE_KEYS = [Qt.Key_Shift, Qt.Key_Control, Qt.Key_Meta, Qt.Key_Alt, Qt.Key_AltGr,
                   Qt.Key_CapsLock, Qt.Key_NumLock, Qt.Key_ScrollLock]
        
    def keyPressEvent(self, event):
        if event.key() in self.IGNORE_KEYS or event.count() != 1:
            return QLabel.keyPressEvent(self, event)
        mod = event.modifiers() & self.SAFE_MODIFIER_MASK
        key = QKeySequence(event.key() | mod).toString().split('+')
        if  event.modifiers() & Qt.KeypadModifier:
            key.insert(-1, 'KP')
        self.key = '+'.join(key)
        self.key_pressed.emit()
        
        
class ShortcutDelegate (QStyledItemDelegate):
    def __init__(self, parent):
        QStyledItemDelegate.__init__(self, parent)
        
    def createEditor(self, parent, unused_option, unused_index):
        editor = ShortcutEditor(parent)
        editor.key_pressed.connect(self.on_editor_key_pressed)
        return editor
        
    def setEditorData(self, editor, index):
        pass
        
    def setModelData(self, editor, model, index):
        if editor.key is not None:
            model.setData(index, editor.key, Qt.DisplayRole)
            
    def on_editor_key_pressed(self):
        editor = self.sender()
        self.commitData.emit(editor)
        self.closeEditor.emit(editor, QAbstractItemDelegate.NoHint)
        
        
class PreferencesDialog (Dialog):
    sample_buffers = 0
        
    def __init__(self, parent):
        Dialog.__init__(self, parent)
        self.ui = UI(self, 'preferences.ui')
        from .ui.preferences import retranslate
        retranslate(self)
        
        self.ui.button_animspeed_reset.setIcon(QIcon.fromTheme('edit-clear'))
        self.ui.button_lighting_reset.setIcon(QIcon.fromTheme('edit-clear'))
        self.ui.button_antialiasing_reset.setIcon(QIcon.fromTheme('edit-clear'))
        self.ui.button_mirror_faces_reset.setIcon(QIcon.fromTheme('edit-clear'))
        self.ui.button_movekey_add.setIcon(QIcon.fromTheme('list-add'))
        self.ui.button_movekey_remove.setIcon(QIcon.fromTheme('list-remove'))
        self.ui.button_movekey_reset.setIcon(QIcon.fromTheme('document-revert'))
        self.ui.button_color_reset.setIcon(QIcon.fromTheme('edit-clear'))
        self.ui.button_image_reset.setIcon(QIcon.fromTheme('edit-clear'))
        self.ui.button_background_color_reset.setIcon(QIcon.fromTheme('edit-clear'))
        self.ui.button_mousemode_quad.setIcon(QIcon(os.path.join(config.UI_DIR, "mousemode-quad.png")))
        self.ui.button_mousemode_ext.setIcon(QIcon(os.path.join(config.UI_DIR, "mousemode-ext.png")))
        
        self.ui.buttonBox.button(QDialogButtonBox.Close).setDefault(True)
        self.ui.label_needs_restarted.setVisible(False)
        
        # graphic tab
        self.ui.slider_animspeed.setValue(settings.draw.speed)
        self.ui.slider_animspeed.setRange(*settings.draw.speed_range)
        self.bind('draw.speed', self.ui.slider_animspeed, 'value', 'setValue', 'valueChanged')
        self.bind_reset('draw.speed', self.ui.button_animspeed_reset)
        
        self.bind('draw.lighting', self.ui.checkbox_lighting, 'isChecked', 'setChecked', 'toggled')
        self.bind_reset('draw.lighting', self.ui.button_lighting_reset)
        
        def reset_antialiasing():
            del settings.draw.samples
        self.ui.button_antialiasing_reset.clicked.connect(reset_antialiasing)
        for text in settings.draw.samples_range:
            self.ui.combobox_samples.addItem(_(text), text)
        self.ui.combobox_samples.setCurrentIndex(settings.draw.samples)
        self.bind('draw.samples', self.ui.combobox_samples,
                                  'currentIndex', 'setCurrentIndex', 'currentIndexChanged')
        
        self.ui.spinbox_mirror_faces.setRange(*settings.draw.mirror_distance_range)
        def set_mirror_faces(checked):
            self.ui.checkbox_mirror_faces.setChecked(checked)
            self.ui.spinbox_mirror_faces.setEnabled(checked)
        self.bind('draw.mirror_faces', self.ui.checkbox_mirror_faces, 'isChecked', set_mirror_faces, 'toggled')
        self.bind_reset('draw.mirror_faces', self.ui.button_mirror_faces_reset)
        self.bind('draw.mirror_distance', self.ui.spinbox_mirror_faces, 'value', 'setValue', 'valueChanged')
        self.bind_reset('draw.mirror_distance', self.ui.button_mirror_faces_reset)
        
        html = '<html><head/><body><p><span style=" color:#aa0000;">{}</span></p></body></html>'
        html = html.format(Qt.escape(_('The program needs to be restarted for the changes to take effect.')))
        self.ui.label_needs_restarted.setText(html)
        
        # mouse tab
        def set_selection_mode(unused_mode):
            if settings.draw.selection_nick == 'quadrant':
                self.ui.button_mousemode_quad.setChecked(True)
            elif settings.draw.selection_nick == 'simple':
                self.ui.button_mousemode_ext.setChecked(True)
        self.bind('draw.selection', None, None, set_selection_mode, None)
        
        # keys tab
        self.liststore_movekeys = QStandardItemModel(self)
        self.fill_liststore_movekeys()
        self.ui.listview_movekeys.setModel(self.liststore_movekeys)
        self.liststore_movekeys.itemChanged.connect(self.on_liststore_movekeys_itemChanged)
        self.shortcut_delegate = ShortcutDelegate(self.ui.listview_movekeys)
        self.ui.listview_movekeys.setItemDelegateForColumn(1, self.shortcut_delegate)
        
        # design tab
        self.image_dirname = os.environ.get('HOME', '') #TODO: Use Picture folder, Qt5 has a class for that
        self.active_face = 0
        self.liststore_faces = self.create_face_selector()
        
        self.button_color = ColorButton(self.ui.button_color)
        self.ui.button_color = None
        self.setTabOrder(self.ui.listview_faces, self.button_color)
        self.button_color.color_changed.connect(self.on_button_color_color_changed)
        self.button_background_color = ColorButton(self.ui.button_background_color)
        self.ui.button_background_color = None
        self.setTabOrder(self.ui.radiobutton_mosaic, self.button_background_color)
        
        # Only a single color, no picture or pattern
        self.ui.combobox_image.addItem(_('plain'), '')
        for filename in textures.stock_files:
            pixmap = QPixmap(textures.get_stock_pixbuf(filename))
            icon = QIcon(pixmap)
            self.ui.combobox_image.addItem(icon, '', filename)
        self.ui.combobox_image.addItem(_('select …'), '/')
        
        self.bind_reset_item('theme.face.{}.color', self.ui.button_color_reset)
        self.bind_reset_item('theme.face.{}.image', self.ui.button_image_reset)
        self.bind('theme.bgcolor', self.button_background_color, 'get_color', 'set_color', 'color_changed')
        self.bind_reset('theme.bgcolor', self.ui.button_background_color_reset)
        
        index = self.liststore_faces.index(self.active_face, 1)
        self.ui.listview_faces.setCurrentIndex(index)
        
    def create_face_selector(self):
        liststore_faces = QStandardItemModel()
        face_names = [_('Up'), _('Down'), _('Left'), _('Right'), _('Front'), _('Back')]
        for i in range(6):
            liststore_faces.appendRow((QStandardItem(i), QStandardItem(face_names[i])))
            filename = settings.theme.face[i].image
            if filename.startswith('/'):
                self.image_dirname = os.path.dirname(filename)
        self.ui.listview_faces.setModel(liststore_faces)
        self.ui.listview_faces.setModelColumn(1)
        #XXX: workaround, listview_faces should automatically set to the correct width
        fm = QFontMetrics(self.ui.listview_faces.font())
        width = max(fm.width(fn) for fn in face_names) + 8
        self.ui.listview_faces.setMaximumWidth(width)
        self.ui.listview_faces.selectionModel().currentChanged.connect(self.on_listview_faces_currentChanged)
        return liststore_faces
        
    def fill_liststore_movekeys(self):
        for move, key in settings.draw.accels:
            self.liststore_movekeys.appendRow([QStandardItem(move), QStandardItem(key)])
        self.liststore_movekeys.setHeaderData(0, Qt.Horizontal, _('Move'))
        self.liststore_movekeys.setHeaderData(1, Qt.Horizontal, _('Key'))
            
    def bind_reset_item(self, settings_key, button):
        def on_clicked():
            delattr(settings, settings_key.format(self.active_face))
        button.clicked.connect(on_clicked)
        
    @ staticmethod
    def _accel_mods_to_str(accel_mods):
        accel_str = ''
        for a in accel_mods.value_nicks:
            if accel_str:
                accel_str += '+'
            if a.endswith('-mask'):
                a = a[:-5]
            accel_str += a
        return accel_str
        
    def set_imagefile(self, imagefile):
        index_icon = len(textures.stock_files) + 1
        if not imagefile:
            index = 0
            icon = QIcon()
        elif imagefile.startswith('/'):
            index = index_icon
            icon = QIcon(imagefile)
        else:
            try:
                index = textures.stock_files.index(imagefile) + 1
            except ValueError:
                index = 0
            icon = QIcon()
        self.ui.combobox_image.setItemIcon(index_icon, icon)
        self.ui.combobox_image.setCurrentIndex(index)
        
    def on_settings_changed(self, key):
        Dialog.on_settings_changed(self, key)
        if self._ignore_changed:
            return
        self._ignore_changed = True
        if key == 'draw.samples':
            visible = (self.sample_buffers != 2**settings.draw.samples > 1)
            self.ui.label_needs_restarted.setVisible(visible)
        elif key == 'draw.accels':
            self.liststore_movekeys.clear()
            self.fill_liststore_movekeys()
        elif key == 'theme.face.{}.color'.format(self.active_face):
            self.button_color.set_color(settings.theme.face[self.active_face].color)
        elif key == 'theme.face.{}.image'.format(self.active_face):
            self.set_imagefile(settings.theme.face[self.active_face].image)
        elif key == 'theme.face.{}.mode':
            imagemode = settings.theme.face[self.active_face].mode_nick
            if imagemode == 'tiled':
                self.ui.radiobutton_tiled.setChecked(True)
            elif imagemode == 'mosaic':
                self.ui.radiobutton_mosaic.setChecked(True)
        self._ignore_changed = False
        
    ###
    
    @Slot(bool)
    def on_checkbox_mirror_faces_toggled(self, checked):
        self.ui.spinbox_mirror_faces.setEnabled(checked)
        
    ### mouse handlers ###
    
    def set_mousemode(self, checked, mode):
        if self._ignore_changed:
            return
        if checked:
            self._ignore_changed = True
            settings.draw.selection_nick = mode
            self._ignore_changed = False
            
    @Slot(bool)
    def on_button_mousemode_quad_toggled(self, checked):
        self.set_mousemode(checked, 'quadrant')
        
    @Slot(bool)
    def on_button_mousemode_ext_toggled(self, checked):
        self.set_mousemode(checked, 'simple')
        
    ### key handlers ###
    
    def get_move_key_list(self):
        move_keys = []
        for i in range(self.liststore_movekeys.rowCount()):
            move, key = [self.liststore_movekeys.item(i, j).data(Qt.DisplayRole) for j in (0, 1)]
            move_keys.append((move, key))
        return move_keys
            
    @Slot()
    def on_button_movekey_add_clicked(self):
        row = self.ui.listview_movekeys.currentIndex().row()
        self._ignore_changed = True
        self.liststore_movekeys.insertRow(row, (QStandardItem(''), QStandardItem('')))
        index = self.liststore_movekeys.index(row, 0)
        self.ui.listview_movekeys.setCurrentIndex(index)
        self.ui.listview_movekeys.edit(index)
        self._ignore_changed = False
        
    @Slot()
    def on_button_movekey_remove_clicked(self):
        row = self.ui.listview_movekeys.currentIndex().row()
        self._ignore_changed = True
        self.liststore_movekeys.takeRow(row)
        settings.draw.accels = self.get_move_key_list()
        self._ignore_changed = False
        
    @Slot()
    def on_button_movekey_reset_clicked(self):  # pylint: disable=R0201
        del settings.draw.accels
        
    def on_liststore_movekeys_itemChanged(self, unused_item):
        self._ignore_changed = True
        settings.draw.accels = self.get_move_key_list()
        self._ignore_changed = False
        
    ### appearance handlers ###
    
    def on_listview_faces_currentChanged(self, current):
        self.active_face = current.row()
        self._ignore_changed = True
        self.button_color.set_color(settings.theme.face[self.active_face].color)
        self.set_imagefile(settings.theme.face[self.active_face].image)
        imagemode = settings.theme.face[self.active_face].mode_nick
        if imagemode == 'tiled':
            self.ui.radiobutton_tiled.setChecked(True)
        elif imagemode == 'mosaic':
            self.ui.radiobutton_mosaic.setChecked(True)
        self._ignore_changed = False
        
    def on_button_color_color_changed(self, color):
        if self._ignore_changed:
            return
        self._ignore_changed = True
        settings.theme.face[self.active_face].color = color
        self._ignore_changed = False
        
    @Slot(int)
    def on_combobox_image_activated(self, index):
        if self._ignore_changed:
            return
        if index == 0:
            filename = ''
        else:
            try:
                filename = textures.stock_files[index-1]
            except IndexError:
                filename = QFileDialog.getOpenFileName(self, _("Open Image"), self.image_dirname)
                if isinstance(filename, (tuple, list)):
                    filename = filename[0]
                if filename == '':
                    # canceled, set the old image
                    filename = settings.theme.face[self.active_face].image
                    self.set_imagefile(filename)
                else:
                    self.image_dirname = os.path.dirname(filename)
        self._ignore_changed = True
        settings.theme.face[self.active_face].image = filename
        self._ignore_changed = False
        
    @Slot(bool)
    def on_radiobutton_tiled_toggled(self, checked):
        self.set_imagemode(checked, 'tiled')
    @Slot(bool)
    def on_radiobutton_mosaic_toggled(self, checked):
        self.set_imagemode(checked, 'mosaic')
        
    def set_imagemode(self, checked, mode):
        if self._ignore_changed:
            return
        if checked:
            self._ignore_changed = True
            settings.theme.face[self.active_face].mode_nick = mode
            self._ignore_changed = False
            
        
class SelectModelDialog (Dialog):
    response_ok = Signal(int, tuple, bool)
    
    def __init__(self, parent):
        Dialog.__init__(self, parent)
        self.ui = UI(self, 'model.ui')
        from .ui.model import retranslate
        retranslate(self)
        
        for text in settings.game.type_range:
            self.ui.combobox_model.addItem(_(text), text)
        mtype = settings.game.type
        self.ui.combobox_model.setCurrentIndex(mtype)
        self.on_combobox_model_activated(mtype)
        size = settings.game.size
        width, height, depth = size
        self.ui.spin_size1.setValue(width)
        self.ui.spin_size2.setValue(height)
        self.ui.spin_size3.setValue(depth)
        
    @Slot(int)
    def on_combobox_model_activated(self, index):
        mtype = models[index].type
        self.ui.label_width.setText({'Cube':_('Size:'), 'Tower':_('Basis:')}.get(mtype, _('Width:')))
        self.ui.label_width.setEnabled(mtype in ['Cube', 'Tower', 'Brick'])
        self.ui.label_heigth.setEnabled(mtype in ['Tower', 'Brick'])
        self.ui.label_depth.setEnabled(mtype == 'Brick')
        self.ui.spin_size1.setEnabled(mtype in ['Cube', 'Tower', 'Brick'])
        self.ui.spin_size2.setEnabled(mtype in ['Tower', 'Brick'])
        self.ui.spin_size3.setEnabled(mtype == 'Brick')
        
    def on_response(self):
        self.response_ok.emit(
                self.ui.combobox_model.currentIndex(),
                (self.ui.spin_size1.value(), self.ui.spin_size2.value(), self.ui.spin_size3.value()),
                self.ui.checkbox_solved.isChecked())
        
        
class ProgressDialog (QProgressDialog):
    def __init__(self, parent):
        QProgressDialog.__init__(self, parent)
        self.canceled_ = False
        self.value_max = 10
        self.value = 0
        self.setWindowModality(Qt.WindowModal)
        self.setMaximum(self.value_max)
        self.setMinimumDuration(0)
        self.setAutoReset(False)
        self.canceled.connect(self.on_canceled)
        
    def on_canceled(self):
        self.canceled_ = True
        self.setLabelText(_('Canceling operation, please wait'))
        
    def tick(self, step, message=None, value_max=None):
        if not self.isVisible():
            self.show()
        if message is not None:
            self.setLabelText(message)
        if value_max is not None:
            self.value_max = value_max
            self.value = 0
            self.setMaximum(value_max)
        if step < 0 or self.value > self.value_max:
            self.setValue(0)
            self.setMaximum(0)
        else:
            self.setValue(self.value)
            self.value += step
        return self.canceled_
        
    def done(self):
        self.canceled_ = False
        self.reset()
        
def linkedtext_to_html(text):
    html = QTextDocumentFragment.fromPlainText(text).toHtml()
    html = re.sub(r'&lt;((?:http:|https:|text:).*?)\|&gt;', r'<a href="\1">', html)
    html = re.sub(r'&lt;\|&gt;', r'</a>', html)
    return re.sub(r'&lt;((?:http:|https:).*?)&gt;', r'&lt;<a href="\1">\1</a>&gt;', html)
    
class UI:
    def __init__(self, toplevel, filename):
        self._toplevel = toplevel
        from PyQt4 import uic
        uic.loadUi(os.path.join(config.UI_DIR, filename), toplevel)
    def __getattr__(self, attrname):
        if attrname[0] == '_':
            return object.__getattr__(self, attrname)
        else:
            return self._toplevel.findChild(QObject, attrname)
            
        
class AboutDialog (QDialog):
    def __init__(self, parent):
        QDialog.__init__(self, parent)
        self.about = UI(self, 'about.ui')
        from .ui.about import retranslate
        retranslate(self)
        
        self.fill_header()
        self.fill_about_tab()
        self.fill_feedback_tab()
        self.fill_translation_tab()
        self.fill_license_tab()
        
        self.index_tab_about = self.about.tab_widget.indexOf(self.about.tab_about)
        self.index_tab_license = self.about.tab_widget.indexOf(self.about.tab_license)
        
        # About tab animation
        self.scrollbar = self.about.text_translators.verticalScrollBar()
        self.animation = QPropertyAnimation(self.scrollbar, 'value')
        self.animation.setLoopCount(-1)
        
    def fill_header(self):
        self.about.label_icon.setPixmap(QPixmap(os.path.join(config.PIXMAP_DIR, 'pybik-64')))
        self.about.label_appname.setText(_(config.APPNAME))
        self.about.label_version.setText(config.VERSION)
        self.about.label_description.setText(_(config.SHORT_DESCRIPTION))
        
    def fill_about_tab(self):
        self.about.label_copyright.setText(config.COPYRIGHT)
        self.about.label_website.setText(
                '<html><head/><body><p><a href="{}">{}</a></p></body></html>'
                .format(config.WEBSITE, _('Pybik project website')))
        html_template = '''<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.0//EN" "http://www.w3.org/TR/REC-html40/strict.dtd">
<html><head><meta name="qrichtext" content="1" /><style type="text/css">
p, li {{ white-space: pre-wrap; }}
</style></head>
<body>{}</body></html>'''
        html_p_template = '<p align="center">{}</p>'
        html_language_template = '<span style=" font-weight:600;">{}</span>'
        html_person_template = '<a href="{}">{}</a>'
        html_p_items = []
        from .translators import translators
        try:
            import icu
        except ImportError:
            print('PyICU is not installed')
            langname_from_code = lambda code, name: name
        else:
            def langname_from_code(code, name):
                iculocale = icu.Locale.createFromName(code)
                lang = iculocale.getLanguage()
                if icu.Locale.createFromName(lang).getDisplayName() == lang:
                    return name
                return str(iculocale.getDisplayName()) or name
        import locale
        sortfunc = lambda v: locale.strxfrm(v[0])
        def gentranslators():
            for language, langcode, persons in translators:
                language = langname_from_code(langcode, language)
                yield language, persons
        for language, persons in sorted(gentranslators(), key=sortfunc):
            language = Qt.escape(language)
            html_items = [html_language_template.format(language)]
            for name, link in sorted(persons, key=sortfunc):
                name = Qt.escape(name)
                html_items.append(html_person_template.format(link, name))
            html_p_items.append(html_p_template.format('<br/>'.join(html_items)))
        html = html_template.format(''.join(html_p_items))
        self.about.text_translators.setHtml(html)
        self.about.text_translators.anchorClicked.connect(QDesktopServices.openUrl)
        self.about.text_translators.viewport().installEventFilter(self)
        
    def fill_feedback_tab(self):
        html = linkedtext_to_html(config.get_filebug_text())
        self.about.label_feedback.setText(html)
        
    def fill_translation_tab(self):
        html = linkedtext_to_html(_(config.TRANSLATION_TEXT))
        self.about.label_translation.setText(html)
        
    def fill_license_tab(self):
        self.about.text_license_short.hide()
        self.about.text_license_full.hide()
        html = linkedtext_to_html('\n\n'.join((_(config.LICENSE_INFO), _(config.LICENSE_FURTHER))))
        self.about.text_license_short.setHtml(html)
        try:
            with open(config.LICENSE_FILE, 'rt', encoding='utf-8') as license_file:
                text = license_file.read()
        except Exception as e:
            print('Unable to find license text:', e)
            text = _(config.LICENSE_NOT_FOUND)
            self.about.text_license_full.setLineWrapMode(QTextEdit.WidgetWidth)
        html = linkedtext_to_html(text)
        self.about.text_license_full.setHtml(html)
        self.about.tab_widget.currentChanged.connect(self._on_tab_widget_currentChanged)
        self.about.text_license_short.anchorClicked.connect(self._on_text_license_anchorClicked)
        
    def update_animation(self):
        smin = self.scrollbar.minimum()
        smax = self.scrollbar.maximum()
        if smax <= smin:
            return
        self.animation.setDuration((smax-smin) * 40)
        self.animation.setKeyValueAt(0., smin)
        self.animation.setKeyValueAt(0.04, smin)
        self.animation.setKeyValueAt(0.50, smax)
        self.animation.setKeyValueAt(0.54, smax)
        self.animation.setKeyValueAt(1., smin)
        
    def showEvent(self, event):
        if self.animation.state() == QAbstractAnimation.Stopped:
            self.update_animation()
            self.scrollbar.hide()
            self.animation.start()
            
    def resizeEvent(self, event):
        self.update_animation()
        
    def eventFilter(self, obj, event):
        #assert obj == self.about.text_translators.viewport()
        if event.type() in [QEvent.MouseButtonPress, QEvent.Wheel]:
            self.animation.pause()
            self.scrollbar.show()
        return False
        
    def _on_tab_widget_currentChanged(self, index):
        if index == self.index_tab_about:
            self.animation.resume()
            self.scrollbar.hide()
        else:
            self.animation.pause()
            self.scrollbar.show()
        if index == self.index_tab_license:
            self.about.text_license_short.setVisible(True)
            self.about.text_license_full.setVisible(False)
        QTimer.singleShot(0, self.adjustSize)
        
    def _on_text_license_anchorClicked(self, url):
        if url.toString() == 'text:FULL_LICENSE_TEXT':
            self.about.text_license_short.setVisible(False)
            self.about.text_license_full.setVisible(True)
        else:
            QDesktopServices.openUrl(url)
            
    def run(self):
        self.exec_()
        self.deleteLater()
        




#-*- coding:utf-8 -*-
# cython: profile=False

#  Pybik -- A 3 dimensional magic cube game.
#  Copyright © 2009-2013  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Ported from GNUbik
# Original filename: cube.c, drwBlock.c
# Original copyright and license:  1998, 2003, 2004  Dale Mellor, John Darrington,  GPL3+

# although this file is compiled with Python3 syntax, Cython needs at least division from __future__
from __future__ import print_function, division

# This line makes cython happy
global __name__, __package__    # pylint: disable=W0604
#px/__compiled = True
__compiled = False

__all__ = ['VERTEX_ATTRIB_LOCATION', 'NORMAL_ATTRIB_LOCATION', 'COLOR_ATTRIB_LOCATION', 'TEXCOORD_ATTRIB_LOCATION',
           'gl_draw_cube', 'gl_pick_cube', 'gl_init_buffers', 'gl_delete_buffers',
           'gl_draw_cube_debug', 'gl_init_object_location']

from .debug import debug  # pylint: disable=W0614,W0401

#pxd/from gl cimport *
from OpenGL.GL import *     # pylint: disable=W0614,W0401
#px/from glm cimport *
from .glm import *


debug('Importing module:', __name__)
debug('  from package:', __package__)
debug('  compiled:', __compiled)


#pxd/cdef enum: #
if True:
    MAX_BLOCKS = 1000
    MAX_FACES = 6
    MAX_FACES_PER_BLOCK = 6
    #pxd=4
    VERTEX_ATTRIB_LOCATION = 0
    NORMAL_ATTRIB_LOCATION = 1
    COLOR_ATTRIB_LOCATION = 2
    TEXCOORD_ATTRIB_LOCATION = 3

#px/cdef struct Block:
class Block:     # pylint: disable=R0903
#px-
    def __init__(self):
        # The position from the centre of the cube,  and the rotation from the
        # 'natural' position.
        #px/mat4 transformation
        self.transformation = mat4()
        
        #px/bint in_motion
        self.in_motion = None
        
        #px/int cnt_faces
        self.cnt_faces = None
        #px/int idx_label[MAX_FACES_PER_BLOCK]
        self.idx_label = [None]*MAX_FACES_PER_BLOCK
        #px/int cnt_label[MAX_FACES_PER_BLOCK]
        self.cnt_label = [None]*MAX_FACES_PER_BLOCK
        #px/int texname[MAX_FACES_PER_BLOCK]
        self.texname = [None]*MAX_FACES_PER_BLOCK
        
        #px/int idx_triangles
        self.idx_triangles = None
        #px/int cnt_triangles
        self.cnt_triangles = None
        
#px/cdef struct Cube:
class cube:     # pylint: disable=W0232, R0903
    #px+unsigned int number_blocks
    #px/Block blocks[MAX_BLOCKS]
    blocks = [Block() for __block in range(MAX_BLOCKS)]
    
    #px+int idx_pick
    #px+int cnt_pick
    #px+int idx_debug
    
    #px+GLuint object_location
    #px+GLuint glbuffer
    
#px+cdef Cube cube

#px/cdef struct Animation_Struct:
class animation:    # pylint: disable=W0232, R0903
    #px+float angle, angle_max
    #px/vec3 rotation
    rotation = [None] * 3
    #px+mat4 rotation_matrix
#px+cdef Animation_Struct animation


def init_module():
    #px+cdef int i
    
    cube.number_blocks = 0
    #cube.blocks
    
    cube.idx_pick = 0
    cube.cnt_pick = 0
    cube.idx_debug = 0
    
    animation.angle = animation.angle_max = 0

def set_transformations(blocks):
    #px+cdef unsigned int b,i,j
    for b in range(cube.number_blocks):
        for i in range(4):
            for j in range(4):
                cube.blocks[b].transformation[i][j] = float(blocks[b][i][j])
        cube.blocks[b].in_motion = False

#px/cpdef set_animation_start(blocks, float angle_max, float axisx, float axisy, float axisz):
def set_animation_start(blocks, angle_max, axisx, axisy, axisz):
    animation.angle = 0.0
    animation.angle_max = angle_max
    animation.rotation = vec3(axisx, axisy, axisz)
    animation.rotation_matrix = mat4(1.)
    #px+cdef int block_id
    #px+cdef bint selected
    for block_id, selected in blocks:
        cube.blocks[block_id].in_motion = selected
    
#px/cpdef set_animation_next(float increment):
def set_animation_next(increment):
    animation.angle -= increment
    animation.rotation_matrix = rotate(animation.angle, animation.rotation)
    return abs(animation.angle) < animation.angle_max

#pxd/cdef void gl_draw_cube():
def gl_draw_cube():
    glEnableVertexAttribArray(VERTEX_ATTRIB_LOCATION)
    glEnableVertexAttribArray(NORMAL_ATTRIB_LOCATION)
    glEnableVertexAttribArray(COLOR_ATTRIB_LOCATION)
    glEnableVertexAttribArray(TEXCOORD_ATTRIB_LOCATION)
    #px+cdef unsigned int i
    #px+cdef mat4 object_matrix
    for i in range(cube.number_blocks):
        if cube.blocks[i].in_motion:
            object_matrix = animation.rotation_matrix * cube.blocks[i].transformation
            #px/glUniformMatrix4fv(cube.object_location, 1, GL_FALSE, &object_matrix[0][0])
            glUniformMatrix4fv(cube.object_location, 1, GL_FALSE, object_matrix)
        else:
            #px/glUniformMatrix4fv(cube.object_location, 1, GL_FALSE, &cube.blocks[i].transformation[0][0])
            glUniformMatrix4fv(cube.object_location, 1, GL_FALSE, cube.blocks[i].transformation)
        _gl_draw_block(i)
    glDisableVertexAttribArray(VERTEX_ATTRIB_LOCATION)
    glDisableVertexAttribArray(NORMAL_ATTRIB_LOCATION)
    glDisableVertexAttribArray(COLOR_ATTRIB_LOCATION)
    glDisableVertexAttribArray(TEXCOORD_ATTRIB_LOCATION)
        
#px/cdef void _gl_draw_block(int block_id):
def _gl_draw_block(block_id):
    #px+cdef int i
    
    glBindTexture(GL_TEXTURE_2D, 0)
    glDrawArrays(GL_TRIANGLES, cube.blocks[block_id].idx_triangles, cube.blocks[block_id].cnt_triangles)
    
    for i in range(cube.blocks[block_id].cnt_faces):
        # render the colors (ie the little sticky labels)
        glBindTexture(GL_TEXTURE_2D, cube.blocks[block_id].texname[i])
        #TODO: Workaround for artifacts on the bottom and left sides of images
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_MIRRORED_REPEAT)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_MIRRORED_REPEAT)
        glDrawArrays(GL_TRIANGLES, cube.blocks[block_id].idx_label[i], cube.blocks[block_id].cnt_label[i])
            
#pxd/cdef void gl_pick_cube():
def gl_pick_cube():
    glEnableVertexAttribArray(VERTEX_ATTRIB_LOCATION)
    glEnableVertexAttribArray(COLOR_ATTRIB_LOCATION)
    glDrawArrays(GL_TRIANGLES, cube.idx_pick, cube.cnt_pick)
    glDisableVertexAttribArray(VERTEX_ATTRIB_LOCATION)
    glDisableVertexAttribArray(COLOR_ATTRIB_LOCATION)
    
#pxd/cdef void gl_init_buffers():
def gl_init_buffers():
    #px/glGenBuffers(1, &cube.glbuffer)
    cube.glbuffer = glGenBuffers(1)
    
#pxd/cdef void gl_delete_buffers():
def gl_delete_buffers():
    glBindBuffer(GL_ARRAY_BUFFER, 0)
    #px/glDeleteBuffers(1, &cube.glbuffer)
    glDeleteBuffers(1, [cube.glbuffer])
    cube.glbuffer = 0
    
#px/cpdef gl_set_data(int nblocks, bytes vertexdata, vertexpointers, vertexinfo, texnames):
def gl_set_data(nblocks, vertexdata, vertexpointers, vertexinfo, texnames):
    cube.number_blocks = nblocks
    assert cube.number_blocks <= MAX_BLOCKS
    
    #### Create the raw GL data ####
    #px+cdef long normalpointer, colorpointer, texpospointer
    #px+cdef unsigned int idx_block, idx_pick, cnt_pick, idx_debug
    normalpointer, colorpointer, texpospointer = vertexpointers
    labelinfos, idx_block, cnts_block, idx_pick, cnt_pick, idx_debug = vertexinfo
    
    #px+cdef unsigned int block, i, idx_label, cnt, faceno
    idx_label = 0
    for block in range(cube.number_blocks):
        cube.blocks[block].cnt_faces = len(labelinfos[block])
        for i, (cnt, faceno) in enumerate(labelinfos[block]):
            cube.blocks[block].idx_label[i] = idx_label
            cube.blocks[block].cnt_label[i] = cnt
            cube.blocks[block].texname[i] = texnames[faceno]
            idx_label += cnt
        cube.blocks[block].idx_triangles = idx_block
        cube.blocks[block].cnt_triangles = cnts_block[block]
        idx_block += cnts_block[block]
        
    cube.idx_pick = idx_pick
    cube.cnt_pick = cnt_pick
    cube.idx_debug = idx_debug
    
    debug('GL data: %d bytes' % len(vertexdata))
    glBindBuffer(GL_ARRAY_BUFFER, cube.glbuffer)
    #px/glBufferData(GL_ARRAY_BUFFER, len(vertexdata), <char*>vertexdata, GL_STATIC_DRAW)
    glBufferData(GL_ARRAY_BUFFER, len(vertexdata), vertexdata, GL_STATIC_DRAW)
    
    #px/glVertexAttribPointer(VERTEX_ATTRIB_LOCATION, 3, GL_FLOAT, GL_FALSE, 0, <void*>0)
    glVertexAttribPointer(VERTEX_ATTRIB_LOCATION, 3, GL_FLOAT, GL_FALSE, 0, 0)
    #px/glVertexAttribPointer(NORMAL_ATTRIB_LOCATION, 3, GL_FLOAT, GL_FALSE, 0, <void*>normalpointer)
    glVertexAttribPointer(NORMAL_ATTRIB_LOCATION, 3, GL_FLOAT, GL_FALSE, 0, normalpointer)
    #px/glVertexAttribPointer(COLOR_ATTRIB_LOCATION, 3, GL_UNSIGNED_BYTE, GL_TRUE, 0, <void*>colorpointer)
    glVertexAttribPointer(COLOR_ATTRIB_LOCATION, 3, GL_UNSIGNED_BYTE, GL_TRUE, 0, colorpointer)
    #px/glVertexAttribPointer(TEXCOORD_ATTRIB_LOCATION, 2, GL_FLOAT, GL_FALSE, 0, <void*>texpospointer)
    glVertexAttribPointer(TEXCOORD_ATTRIB_LOCATION, 2, GL_FLOAT, GL_FALSE, 0, texpospointer)
    
#pxd/cdef void gl_draw_cube_debug():
def gl_draw_cube_debug():
    #px+cdef mat4 object_matrix
    object_matrix = mat4(1.)
    #px/glUniformMatrix4fv(cube.object_location, 1, GL_FALSE, &object_matrix[0][0])
    glUniformMatrix4fv(cube.object_location, 1, GL_FALSE, object_matrix)
    glEnableVertexAttribArray(VERTEX_ATTRIB_LOCATION)
    glEnableVertexAttribArray(COLOR_ATTRIB_LOCATION)
    glDrawArrays(GL_LINES, cube.idx_debug, 6)
    glDisableVertexAttribArray(VERTEX_ATTRIB_LOCATION)
    glDisableVertexAttribArray(COLOR_ATTRIB_LOCATION)
    
#pxd/cdef void gl_draw_select_debug(GLfloat *selectdata, GLsizeiptr size, GLuint prog_hud):
def gl_draw_select_debug(selectdata, size, prog_hud):
    #px+cdef int i, j
    #px+cdef GLintptr offset
    offset = (cube.idx_debug+6) * 3 * sizeof(GLfloat)
    #px/glBufferSubData(GL_ARRAY_BUFFER, offset, size, &selectdata[0])
    glBufferSubData(GL_ARRAY_BUFFER, offset, size, selectdata)
    
    glDisable(GL_DEPTH_TEST)
    glEnableVertexAttribArray(VERTEX_ATTRIB_LOCATION)
    glEnableVertexAttribArray(COLOR_ATTRIB_LOCATION)
    glDrawArrays(GL_LINES, cube.idx_debug+6, 2)
    glUseProgram(prog_hud)
    glDrawArrays(GL_POINTS, cube.idx_debug+6+2, 2)
    glDisableVertexAttribArray(VERTEX_ATTRIB_LOCATION)
    glDisableVertexAttribArray(COLOR_ATTRIB_LOCATION)
    glEnable(GL_DEPTH_TEST)
    
#pxd/cdef void gl_init_object_location(GLuint location):
def gl_init_object_location(location):
    cube.object_location = location
    




#-*- coding:utf-8 -*-

#  Pybik -- A 3 dimensional magic cube game.
#  Copyright © 2009-2013  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Ported from GNUbik
# Original filename: ui.c
# Original copyright and license: GPL3+
#                  1998,  2003  John Darrington
#                  2004  John Darrington,  Dale Mellor



from .debug import *    # pylint: disable=W0614,W0401

from .model import empty_model
from . import cubestate
from .moves import MoveQueue


class GameState (object):
    def __init__(self):
        self.mark_before = False
        self.initial_cube_state = cubestate.CubeState(empty_model)
        self.initial_cube_state.set_solved()
        self.current_cube_state = self.initial_cube_state.copy()
        self.all_moves = MoveQueue()
        
    def copy(self):
        gamestate = GameState()
        gamestate.mark_before = self.mark_before
        gamestate.initial_cube_state = self.initial_cube_state.copy()
        gamestate.current_cube_state = self.current_cube_state.copy()
        gamestate.all_moves = self.all_moves.copy()
        return gamestate
        
    def set_state(self, model, blocks, moves, position):
        self.initial_cube_state = cubestate.CubeState(model)
        try:
            self.initial_cube_state.parse_block(blocks)
        except ValueError as e:
            debug('Invalid block list:', *e.args)
            debug('  start with solved game')
            self.initial_cube_state.set_solved()
        self.current_cube_state = self.initial_cube_state.copy()
        self.all_moves = MoveQueue()
        if moves.startswith('native:'):
            moves = moves.split(':', 1)[1].lstrip()
        else:
            moves = ''
        pos, unused_cpos = self.all_moves.parse(moves, position, model)
        self.do_fast_forward(to_pos=pos)
        
    def get_state(self):
        model = self.initial_cube_state.model
        blocks = self.initial_cube_state.format_block()
        moves, position, unused_markup = self.all_moves.format(model)
        return model, blocks, 'native: ' + moves, position
        
    def set_model(self, model):
        self.initial_cube_state.set_model(model)
        self.current_cube_state.set_model(model)
        
    def new_game(self, model, solved):
        self.initial_cube_state = cubestate.CubeState(model)
        self.initial_cube_state.set_solved()
        if not solved:
            self.initial_cube_state.randomize()
        self.current_cube_state = self.initial_cube_state.copy()
        self.all_moves.reset()
        
    ### Funktions to control the cube for use in callbacks and plugins
    
    def request_back(self):
        '''One step back in the sequence of moves'''
        if self.all_moves.at_start():
            return None
        self.mark_before = self.all_moves.is_mark_current()
        self.all_moves.retard()
        move = self.all_moves.current().inverted()
        self.current_cube_state.rotate_slice(move)
        return move
        
    def request_next(self):
        '''One step forward in the sequence of moves'''
        move = self.all_moves.current()
        if move:
            self.mark_before = self.all_moves.is_mark_current()
            self.current_cube_state.rotate_slice(move)
            self.all_moves.advance()
        return move
        
    def request_rotation(self, maxis, mslice, mdir):
        '''Make one new move.
        
        The usual rotation request,  called when user uses mouse to rotate the cube by
        hand. The move is put into the `current' place,  and all the moves in
        all_moves afterwards are zapped.'''
        self.all_moves.push_current((maxis, mslice, mdir))
        
    def do_rewind_to_mark(self):
        if self.all_moves.at_start():
            return False
        while True:
            self.all_moves.retard()
            move = self.all_moves.current().inverted()
            self.current_cube_state.rotate_slice(move)
            if self.all_moves.is_mark_current():
                break
        return True
        
    def do_rewind_to_start(self):
        # for edit mode cloning is a must, even if already at start
        self.current_cube_state = self.initial_cube_state.copy()
        self.all_moves.rewind_start()
            
    def do_fast_forward(self, to_pos=-1):
        '''Go to end of the sequence of moves.
        
        All the moves take place on the cube,  not via the animation
        routines,  so the effect is to show the result instantaneously.'''
        move = self.all_moves.current()
        while move:
            if to_pos == self.all_moves.current_place:
                break
            self.current_cube_state.rotate_slice(move)
            self.all_moves.advance()
            if to_pos < 0 and self.all_moves.is_mark_current():
                break
            move = self.all_moves.current()
        
    def set_from_formula(self, code, pos):
        self.current_cube_state = self.initial_cube_state.copy()
        self.all_moves.rewind_start()
        self.all_moves.reset()
        pos, unused_cpos = self.all_moves.parse(code, pos, self.current_cube_state.model)
        self.do_fast_forward(to_pos=pos)
        
    def set_as_initial_state(self):
        self.initial_cube_state = self.current_cube_state.copy()
        self.all_moves.truncate_before()
        
    def invert_moves(self):
        self.current_cube_state = self.initial_cube_state.copy()
        self.all_moves.invert()
        pos = self.all_moves.current_place
        self.all_moves.rewind_start()
        self.do_fast_forward(to_pos=pos)
        
    def normalize_complete_rotations(self):
        self.current_cube_state = self.initial_cube_state.copy()
        self.all_moves.normalize_complete_rotations(self.initial_cube_state.model)
        pos = self.all_moves.current_place
        self.all_moves.rewind_start()
        self.do_fast_forward(to_pos=pos)
        
    def swap_block(self, blockpos, maxis, mslice, mdir):
        self.initial_cube_state.swap_block(blockpos, maxis, mslice, mdir)
        # Because initial_cube_state changed, current_cube_state must be cloned now,
        # do_rewind_to_start does that.
        self.do_rewind_to_start()
        
    def rotate_block(self, blockpos, rdir):
        self.initial_cube_state.rotate_block(blockpos, rdir)
        self.do_rewind_to_start()
        




#!/usr/bin/python3
# -*- coding: utf-8 -*-

#  Copyright © 2012-2013  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.



import sys, os
sys.path.insert(0, '.')
import pickle as pickle
from multiprocessing import Pool, cpu_count

from pybiklib import model
model.use_modeldata = False


dedup_data = []
cnt_dups = []
cnt_values = []

def dedup_reset():
    dedup_data[:] = [[] for unused_i in range(6)]
    cnt_dups[:] = [0] * 6
    cnt_values[:] = [0] * 6
    
def pool_dedup(level, value):
    if isinstance(value, dict):
        iterator = list(value.items())
    elif isinstance(value, list):
        iterator = enumerate(value)
    elif isinstance(value, tuple):
        iterator = []
    else:
        return value
    cnt_values[level] += 1
    try:
        didx = dedup_data[level].index((type(value), value))
    except ValueError:
        dedup_data[level].append((type(value), value))
        for k, v in iterator:
            v = pool_dedup(level+1, v)
            value[k] = v
    else:
        dtype, dvalue = dedup_data[level][didx]
        assert value == dvalue and type(value) == type(dvalue) == dtype, (value, dvalue)
        cnt_dups[level] += 1
        value = dvalue
    return value
        
def pool_functions(parallel):
    pool = None
    try:
        if parallel > 1:
            pool = Pool(processes=parallel)
    except OSError as e:
        print('process pool not available ({}):'.format(e))
        print('  maybe /dev/shm is not mounted,')
        print('  deactivating multiprocessing')
    sys.stdout.flush() # when Pool(…) fails this line is 
    if pool is not None:
        return pool.imap_unordered
    else:
        return map
        
def enum_modelfiles(dirname, testfunc):
    ignored = []
    modelfiles = []
    for Model in model.models:
        for i in range(1, 11):
            for j in range(1, 11):
                for k in range(1, 11):
                    sizes = Model.norm_sizes((i, j, k))
                    filename = Model.get_datafilename(sizes)
                    filename = os.path.join(dirname, filename)
                    if filename in modelfiles or filename in ignored:
                        continue
                    if testfunc(filename):
                        modelfiles.append(filename)
                    else:
                        ignored.append(filename)
    if modelfiles:
        for filename in sorted(ignored):
            print('skipping', filename)
    return sorted(modelfiles, reverse=True)
    
def pool_enum_modelfile(filename):
    sizes_list = []
    for Model in model.models:
        for i in range(1, 11):
            for j in range(1, 11):
                for k in range(1, 11):
                    sizes = Model.norm_sizes((i, j, k))
                    if os.path.basename(filename) != Model.get_datafilename(sizes):
                        continue
                    if (Model, sizes) in sizes_list:
                        continue
                    sizes_list.append((Model, sizes))
    return sizes_list
    
def pool_create_modelfiledata(sizes_list):
    savedata = {}
    for Model, sizes in sizes_list:
        modelobj = Model(sizes, None)
        savedata[(Model.type, sizes)] = modelobj.get_savedata()
    return savedata
    
def pool_create_modelfile(args):
    filename, fast = args
    sizes_list = pool_enum_modelfile(filename)
    savedata = pool_create_modelfiledata(sizes_list)
    if not fast:
        dedup_reset()
        savedata = pool_dedup(0, savedata)
    with open(filename, 'wb') as datafile:
        pickle.dump(savedata, datafile, -1)
    vals = ['%6s' % v for v in cnt_values]
    dups = ['%6s' % d for d in cnt_dups]
    return 'generated {} with {} models\n  vals: {}\n  dups: {}'.format(
                    filename, len(sizes_list), ' '.join(vals), ' '.join(dups))
    
def create_modeldata(dirname, testfunc=None, parallel=1, fast=False):
    if testfunc is None:
        testfunc = lambda arg: True
    imap_model = pool_functions(parallel)
    modelfiles = enum_modelfiles(dirname, testfunc)
    for lines in imap_model(pool_create_modelfile, [(m, fast) for m in modelfiles]):
        print(lines)
        sys.stdout.flush()
    
    
if __name__ == '__main__':
    create_modeldata('data/models', parallel=cpu_count())
    #create_modeldata('data/models', parallel=1)
    




#!/usr/bin/python3
# -*- coding: utf-8 -*-

#  Copyright © 2012-2013  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.



import sys, os
sys.path.insert(0, '.')
import re

from pybiklib import config
from pybiklib.config import wrap

try:
    _
except NameError:
    import builtins
    builtins.__dict__['_'] = lambda s: s


def underline(title, underline_char='='):
    return '{}\n{}'.format(title, underline_char*len(title))

def clean_links(text):
    footnotes = []
    def add_footnote(match):
        if match.group(2) != '|':
            return match.group(0)
        elif match.group(1) != '':
            footnote = match.group(1)
            if footnote in footnotes:
                footnotes.append(footnotes.index(footnote)+1)
            else:
                footnotes.append(footnote)
            return ''
        else:
            footnote = footnotes[-1]
            if type(footnote) is int:
                del footnotes[-1]
                return ' [{}]'.format(footnote)
            else:
                return ' [{}]'.format(len(footnotes))
    text = re.sub(r'<(.*?)(\|?)>', add_footnote, text)
    return text, ['[{}] <{}>'.format(i+1, footnote) for i, footnote in enumerate(footnotes)]
    
def read_utf8(filename):
    with open(filename, 'rt', encoding='utf-8') as f:
        return f.read()
    
def write_utf8(filename, text):
    with open(filename, 'wt', encoding='utf-8') as f:
        f.write(text)
        
def create_README(filename=None, skip_install=False):
    if not filename:
        filename = 'README'
    news_template = [
            ('{appname}',   '{description}\n\n'
                            'Author: {AUTHOR} <{CONTACT_EMAIL}>\n'
                            'License: {LICENSE_NAME}\n'
                            'Project page: <{WEBSITE}>'),
            ('Installation', wrap(
                            'If Pybik is available for your distribution, '
                            'you should install Pybik with a package manager. '
                            'Otherwise see the file INSTALL for installation instructions.')),
            ('Feedback',    '{feedback}'),
            ('Translation', '{translation}'),
            ('License', '{license}'),
        ]
    readme = {
            'appname':      'About {} {}'.format(config.APPNAME, config.VERSION),
            'description':  wrap(config.LONG_DESCRIPTION),
            'AUTHOR':       config.AUTHOR,
            'CONTACT_EMAIL':config.CONTACT_EMAIL,
            'LICENSE_NAME': config.LICENSE_NAME,
            'WEBSITE':      config.WEBSITE,
            'feedback':     wrap(clean_links(config.get_filebug_text())),
            'translation':  wrap(clean_links(config.TRANSLATION_TEXT)),
            'license':      wrap(config.LICENSE_INFO) + '\n\n' + wrap(config.LICENSE_NOT_FOUND),
        }
        
    text = ''
    for title, body in news_template:
        if skip_install and title == 'Installation':
            continue
        title = underline(title.format(**readme))
        body = body.format(**readme)
        text += '\n{}\n\n{}\n\n'.format(title, body)
    write_utf8(filename, text)
    
    
def main():
    skip_install = False
    args = sys.argv[1:]
    if not args:
        print('usage: {} [--skip-install] README[=filename]'.format(os.path.basename(sys.argv[0])))
        sys.exit(1)
    for arg in args:
        if arg == '--skip-install':
            skip_install = True
        elif not arg.startswith('--'):
            fileinfo = arg.split('=', 1)
            if len(fileinfo) == 1:
                fileinfo.append(None)
            if fileinfo[0] == 'README':
                create_README(fileinfo[1], skip_install=skip_install)
            else:
                print('Unknown Argument:', arg)
                sys.exit(1)
        else:
            print('Unknown Option:', arg)
            sys.exit(1)
            
    
if __name__ == '__main__':
    try:
        _
    except NameError:
        __builtins__._ = lambda text: text
    main()
    




# -*- coding: utf-8 -*-

#  Copyright © 2013  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

import math
import numpy


class vec3 (numpy.ndarray):
    def __new__(cls, x=0., y=0., z=0.):
        buf = numpy.array([x, y, z], dtype='f')
        return numpy.ndarray.__new__(cls, (3,), dtype='f', buffer=buf)
        
    def _seti(self, i, value):
        self[i] = value
        return value
    x = property(lambda self: self[0], lambda self, value: self._seti(0, value))
    y = property(lambda self: self[1], lambda self, value: self._seti(1, value))
    z = property(lambda self: self[2], lambda self, value: self._seti(2, value))
    
    
class vec4 (numpy.ndarray):
    def __new__(cls, x=0., y=0., z=0., w=1.):
        buf = numpy.array([x, y, z, w], dtype='f')
        return numpy.ndarray.__new__(cls, (4,), dtype='f', buffer=buf)
        
    def _seti(self, i, value):
        self[i] = value
        return value
    x = property(lambda self: self[0], lambda self, value: self._seti(0, value))
    y = property(lambda self: self[1], lambda self, value: self._seti(1, value))
    z = property(lambda self: self[2], lambda self, value: self._seti(2, value))
    w = property(lambda self: self[3], lambda self, value: self._seti(3, value))
    
    
class mat4(numpy.ndarray):
    def __new__(cls, diag=1.):
        buf = numpy.array([[diag, 0., 0., 0.],
                           [0., diag, 0., 0.],
                           [0., 0., diag, 0.],
                           [0., 0., 0., 1.]], dtype='f')
        return numpy.ndarray.__new__(cls, (4,4), dtype='f', buffer=buf)
        
        
def rotate(angle, vect):
    result = mat4()
    x = vect[0]
    y = vect[1]
    z = vect[2]
    c = math.cos(angle)
    s = math.sin(angle)
    result[0][0] = x*x*(1-c) + c
    result[0][1] = x*y*(1-c) - z*s
    result[0][2] = x*z*(1-c) + y*s
    result[1][0] = y*x*(1-c) + z*s
    result[1][1] = y*y*(1-c) + c
    result[1][2] = y*z*(1-c) - x*s
    result[2][0] = x*z*(1-c) - y*s
    result[2][1] = y*z*(1-c) + x*s
    result[2][2] = z*z*(1-c) + c
    return result
    



#-*- coding:utf-8 -*-

#  Copyright © 2009-2013  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


random = lambda game: game.random()
random1 = lambda game: game.random(1)
random2 = lambda game: game.random(2)
random3 = lambda game: game.random(3)
random4 = lambda game: game.random(4)
random5 = lambda game: game.random(5)
random6 = lambda game: game.random(6)
random7 = lambda game: game.random(7)
random8 = lambda game: game.random(8)
random9 = lambda game: game.random(9)
random10 = lambda game: game.random(10)




#!/usr/bin/python3
# -*- coding: utf-8 -*-

#  Copyright © 2011-2013  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


import sys, os, re

def convert(source_filename, target_filename):
    with open(source_filename, 'rt', encoding='utf-8') as source:
        lines = source.readlines()
    re_field = re.compile(r'^(.*): (.+)$')
    with open(target_filename, 'wt', encoding='utf-8') as target:
        translation_comment = None
        dstlineno = 0
        for srclineno, line in enumerate(lines):
            match = re_field.match(line)
            field, value = match.groups() if match else (None, None)
            if field == 'Path':
                sep = value[0]
                formatstr = '_({!r})' if sep != '@' else '{!r}'
                values = []
                for v in value.split(sep):
                    if v:
                        if translation_comment and translation_key in re.split('\W+', v):
                            values.append(translation_comment)
                        v = formatstr.format(v)
                        values.append(v)
                value = '\n'.join(values)
                dstlineno += len(values)
                translation_comment = None
            elif line.strip().startswith('#'):
                translation_comment = line.strip()
                translation_key = translation_comment.strip('#').strip().split()
                if translation_key:
                    translation_key = translation_key[0]
                else:
                    translation_comment = None
                value = None
            else:
                translation_comment = None
                value = None
            if value is None:
                if dstlineno <= srclineno:
                    value = ''
                    dstlineno += 1
                else:
                    continue
            target.write(value)
            target.write('\n')
    
def main():
    if len(sys.argv) != 2:
        print('usage: {} scriptfile'.format(os.path.basename(sys.argv[0])))
        return 1
    filename = sys.argv[1]
    convert(filename, filename+'.py')
    
if __name__ == '__main__':
    sys.exit(main())


